##############################################################################
#
# Copyright (c) 2023 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import pprint
import time
from operator import itemgetter

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import Implicit
from App.special_dtml import DTMLFile
from OFS.SimpleItem import Item


class ZODBConnectionDebugger(Item, Implicit):
    id = 'ZODBConnectionDebugger'
    name = title = 'ZODB Connections'
    meta_type = 'ZODB Connection Debugger'
    zmi_icon = 'fas fa-bug'

    security = ClassSecurityInfo()

    manage_zodb_conns = manage_main = manage = manage_workspace = DTMLFile(
        'dtml/zodbConnections', globals())
    manage_zodb_conns._setName('manage_zodb_conns')
    manage_options = (
        {'label': 'Control Panel', 'action': '../manage_main'},
        {'label': 'Databases', 'action': '../Database/manage_main'},
        {'label': 'Configuration', 'action': '../Configuration/manage_main'},
        {'label': 'DAV Locks', 'action': '../DavLocks/manage_main'},
        {'label': 'Reference Counts', 'action': '../DebugInfo/manage_main'},
        {'label': 'ZODB Connections', 'action': 'manage_main'},
    )

    def dbconnections(self):
        import Zope2  # for data

        result = []
        now = time.time()

        def get_info(connection):
            # `result`, `time` and `before` are lexically inherited.
            request_info = {}
            request_info_formatted = ''
            debug_info_formatted = ''
            opened = connection.opened
            debug_info = connection.getDebugInfo() or {}

            if debug_info:
                debug_info_formatted = pprint.pformat(debug_info)
                if len(debug_info) == 2:
                    request_info = debug_info[0]
                    request_info.update(debug_info[1])
                    request_info_formatted = pprint.pformat(request_info)

            if opened is not None:
                # output UTC time with the standard Z time zone indicator
                open_since = "{}".format(
                    time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(opened)))
                open_for = f"{now - opened:.2f}s"
            else:
                open_since = '(closed)'
                open_for = ''

            result.append({
                'open_since': open_since,
                'open_for': open_for,
                'info': debug_info,
                'info_formatted': debug_info_formatted,
                'request_info': request_info,
                'request_formatted': request_info_formatted,
                'before': connection.before,
                'cache_size': len(connection._cache),
            })

        Zope2.DB._connectionMap(get_info)
        return sorted(result, key=itemgetter('open_since'))


InitializeClass(ZODBConnectionDebugger)



import os
import sys
import time
import unittest
from io import BytesIO

import OFS.Image
import Testing.testbrowser
import Testing.ZopeTestCase
import transaction
import Zope2
from Acquisition import aq_base
from OFS.Application import Application
from OFS.Cache import ZCM_MANAGERS
from OFS.Image import Pdata
from OFS.SimpleItem import SimpleItem
from Testing.makerequest import makerequest
from zExceptions import Redirect
from zope.component import adapter
from zope.datetime import rfc1123_date
from zope.lifecycleevent.interfaces import IObjectCreatedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse


here = os.path.dirname(os.path.abspath(__file__))
filedata = os.path.join(here, 'test.gif')

Zope2.startup_wsgi()


def makeConnection():
    import ZODB
    from ZODB.DemoStorage import DemoStorage

    s = DemoStorage()
    return ZODB.DB(s).open()


def aputrequest(file, content_type):
    resp = HTTPResponse(stdout=sys.stdout)
    environ = {}
    environ['SERVER_NAME'] = 'foo'
    environ['SERVER_PORT'] = '80'
    environ['REQUEST_METHOD'] = 'PUT'
    environ['CONTENT_TYPE'] = content_type
    req = HTTPRequest(stdin=file, environ=environ, response=resp)
    return req


class DummyCache:

    def __init__(self):
        self.clear()

    def ZCache_set(self, ob, data, view_name='', keywords=None,
                   mtime_func=None):
        self.set = (ob, data)

    def ZCache_get(self, ob, data, view_name='', keywords=None,
                   mtime_func=None):
        self.get = ob
        if self.si:
            return self.si

    def ZCache_invalidate(self, ob):
        self.invalidated = ob

    def clear(self):
        self.set = None
        self.get = None
        self.invalidated = None
        self.si = None

    def setStreamIterator(self, si):
        self.si = si


ADummyCache = DummyCache()


class DummyCacheManager(SimpleItem):
    def ZCacheManager_getCache(self):
        return ADummyCache


class EventCatcher:

    def __init__(self):
        self.created = []
        self.modified = []
        self.setUp()

    def setUp(self):
        from zope.component import provideHandler
        provideHandler(self.handleCreated)
        provideHandler(self.handleModified)

    def tearDown(self):
        from zope.component import getSiteManager
        getSiteManager().unregisterHandler(self.handleCreated)
        getSiteManager().unregisterHandler(self.handleModified)

    def reset(self):
        self.created = []
        self.modified = []

    @adapter(IObjectCreatedEvent)
    def handleCreated(self, event):
        if isinstance(event.object, OFS.Image.File):
            self.created.append(event)

    @adapter(IObjectModifiedEvent)
    def handleModified(self, event):
        if isinstance(event.object, OFS.Image.File):
            self.modified.append(event)


class FileTests(unittest.TestCase):
    content_type = 'application/octet-stream'
    factory = 'manage_addFile'

    def setUp(self):
        with open(filedata, 'rb') as fd:
            self.data = fd.read()
        self.connection = makeConnection()
        self.eventCatcher = EventCatcher()
        try:
            r = self.connection.root()
            a = Application()
            r['Application'] = a
            self.root = a
            responseOut = self.responseOut = BytesIO()
            self.app = makerequest(self.root, stdout=responseOut)
            self.app.dcm = DummyCacheManager()
            factory = getattr(self.app, self.factory)
            factory('file',
                    file=self.data, content_type=self.content_type)
            self.app.file.ZCacheable_setManagerId('dcm')
            self.app.file.ZCacheable_setEnabled(enabled=1)
            setattr(self.app, ZCM_MANAGERS, ('dcm',))
            # Hack, we need a _p_mtime for the file, so we make sure that it
            # has one.
            transaction.commit()
        except Exception:
            transaction.abort()
            self.connection.close()
            raise
        transaction.begin()
        self.file = getattr(self.app, 'file')

        # Since we do the create here, let's test the events here too
        self.assertEqual(1, len(self.eventCatcher.created))
        self.assertIs(
            aq_base(self.eventCatcher.created[0].object),
            aq_base(self.file)
        )

        self.assertEqual(1, len(self.eventCatcher.modified))
        self.assertIs(
            aq_base(self.eventCatcher.created[0].object),
            aq_base(self.file)
        )

        self.eventCatcher.reset()

    def tearDown(self):
        del self.file
        transaction.abort()
        self.connection.close()
        del self.app
        del self.responseOut
        del self.root
        del self.connection
        ADummyCache.clear()
        self.eventCatcher.tearDown()

    def testViewImageOrFile(self):
        self.assertRaises(Redirect, self.file.view_image_or_file, 'foo')

    def testUpdateData(self):
        self.file.update_data(b'foo')
        self.assertEqual(self.file.size, 3)
        self.assertEqual(self.file.data, b'foo')
        self.assertTrue(ADummyCache.invalidated)
        self.assertTrue(ADummyCache.set)

    def testReadData(self):
        s = b'a' * (2 << 16)
        data, size = self.file._read_data(BytesIO(s))
        self.assertIsInstance(data, Pdata)
        self.assertEqual(bytes(data), s)
        self.assertEqual(len(s), len(bytes(data)))
        self.assertEqual(len(s), size)

    def testBigPdata(self):
        # Test that a big enough string is split into several Pdata
        # From a file
        s = b'a' * (1 << 16) * 3
        data, size = self.file._read_data(BytesIO(s))
        self.assertNotEqual(data.next, None)
        # From a string
        data, size = self.file._read_data(s)
        self.assertNotEqual(data.next, None)

    def testManageEditWithFileData(self):
        self.file.manage_edit('foobar', 'text/plain', filedata=b'ASD')
        self.assertEqual(self.file.title, 'foobar')
        self.assertEqual(self.file.content_type, 'text/plain')
        self.assertTrue(ADummyCache.invalidated)
        self.assertTrue(ADummyCache.set)
        self.assertEqual(1, len(self.eventCatcher.modified))
        self.assertIs(self.eventCatcher.modified[0].object, self.file)

    def testManageEditWithoutFileData(self):
        self.file.manage_edit('foobar', 'text/plain')
        self.assertEqual(self.file.title, 'foobar')
        self.assertEqual(self.file.content_type, 'text/plain')
        self.assertTrue(ADummyCache.invalidated)
        self.assertEqual(1, len(self.eventCatcher.modified))
        self.assertIs(self.eventCatcher.modified[0].object, self.file)

    def testManageUpload(self):
        f = BytesIO(b'jammyjohnson')
        self.file.manage_upload(f)
        self.assertEqual(self.file.data, b'jammyjohnson')
        self.assertEqual(self.file.content_type, 'application/octet-stream')
        self.assertEqual(1, len(self.eventCatcher.modified))
        self.assertIs(self.eventCatcher.modified[0].object, self.file)

    def testManageUploadWithoutFileData(self):
        self.file.manage_upload()
        self.assertEqual(0, len(self.eventCatcher.modified))

    def testIfModSince(self):
        now = time.time()
        e = {'SERVER_NAME': 'foo',
             'SERVER_PORT': '80',
             'REQUEST_METHOD': 'GET'}

        # not modified since
        t_notmod = rfc1123_date(now)
        e['HTTP_IF_MODIFIED_SINCE'] = t_notmod
        out = BytesIO()
        resp = HTTPResponse(stdout=out)
        req = HTTPRequest(sys.stdin, e, resp)
        data = self.file.index_html(req, resp)
        self.assertEqual(resp.getStatus(), 304)
        self.assertEqual(data, b'')

        # modified since
        t_mod = rfc1123_date(now - 100)
        e['HTTP_IF_MODIFIED_SINCE'] = t_mod
        out = BytesIO()
        resp = HTTPResponse(stdout=out)
        req = HTTPRequest(sys.stdin, e, resp)
        data = self.file.index_html(req, resp)
        self.assertEqual(resp.getStatus(), 200)
        self.assertEqual(data, bytes(self.file.data))

    def testPUT(self):
        s = b'# some python\n'

        # with content type
        data = BytesIO(s)
        req = aputrequest(data, 'text/x-python')
        req.processInputs()
        self.file.PUT(req, req.RESPONSE)

        self.assertEqual(self.file.content_type, 'text/x-python')
        self.assertEqual(self.file.data, s)

        # without content type
        data.seek(0)
        req = aputrequest(data, '')
        req.processInputs()
        self.file.PUT(req, req.RESPONSE)

        self.assertEqual(self.file.content_type, 'text/x-python')
        self.assertEqual(self.file.data, s)

    def testIndexHtmlWithPdata(self):
        self.file.manage_upload(b'a' * (2 << 16))  # 128K
        self.file.index_html(self.app.REQUEST, self.app.REQUEST.RESPONSE)
        self.assertTrue(self.app.REQUEST.RESPONSE._wrote)

    def testIndexHtmlWithString(self):
        self.file.manage_upload(b'a' * 100)  # 100 bytes
        self.file.index_html(self.app.REQUEST, self.app.REQUEST.RESPONSE)
        self.assertTrue(not self.app.REQUEST.RESPONSE._wrote)

    def testPrincipiaSearchSource_not_text(self):
        data = ''.join([chr(x) for x in range(256)])
        data = data.encode('utf-8')
        self.file.manage_edit('foobar', 'application/octet-stream',
                              filedata=data)
        self.assertEqual(self.file.PrincipiaSearchSource(), b'')

    def testPrincipiaSearchSource_text(self):
        self.file.manage_edit('foobar', 'text/plain',
                              filedata=b'Now is the time for all good men to '
                                       b'come to the aid of the Party.')
        self.assertIn(b'Party', self.file.PrincipiaSearchSource())

    def test_manage_DAVget_binary(self):
        self.assertEqual(self.file.manage_DAVget(), self.data)

    def test_manage_DAVget_text(self):
        text = (b'Now is the time for all good men to '
                b'come to the aid of the Party.')
        self.file.manage_edit('foobar', 'text/plain', filedata=text)
        self.assertEqual(self.file.manage_DAVget(), text)

    def test_interfaces(self):
        from OFS.Image import File
        from OFS.interfaces import IWriteLock
        from zope.interface.verify import verifyClass
        from ZPublisher.HTTPRangeSupport import HTTPRangeInterface

        verifyClass(HTTPRangeInterface, File)
        verifyClass(IWriteLock, File)

    def testUnicode(self):
        val = 'some unicode string here'

        self.assertRaises(TypeError, self.file.update_data,
                          data=val, content_type='text/plain')

    def test__str__returns_native_string(self):
        small_data = b'small data'
        self.file.manage_upload(file=small_data)
        self.assertEqual(str(self.file), small_data.decode())

        # Make sure Pdata contents are handled correctly
        big_data = b'a' * (2 << 16)
        self.file.manage_upload(file=big_data)
        self.assertEqual(str(self.file), big_data.decode())


class ImageTests(FileTests):
    content_type = 'image/gif'
    factory = 'manage_addImage'

    def testUpdateData(self):
        self.file.update_data(self.data)
        self.assertEqual(self.file.size, len(self.data))
        self.assertEqual(self.file.data, self.data)
        self.assertEqual(self.file.width, 16)
        self.assertEqual(self.file.height, 16)
        self.assertTrue(ADummyCache.invalidated)
        self.assertTrue(ADummyCache.set)

    def testTag(self):
        tag_fmt = ('<img src="http://nohost/file" '
                   'alt="%s" title="%s" height="16" width="16" />')
        self.assertEqual(self.file.tag(), (tag_fmt % ('', '')))
        self.file.manage_changeProperties(title='foo')
        self.assertEqual(self.file.tag(), (tag_fmt % ('', 'foo')))
        self.file.manage_changeProperties(alt='bar')
        self.assertEqual(self.file.tag(), (tag_fmt % ('bar', 'foo')))

    testStr = testTag

    def test__str__returns_native_string(self):
        small_data = b'small data'
        self.file.manage_upload(file=small_data)
        self.assertIsInstance(str(self.file), str)

        # Make sure Pdata contents are handled correctly
        big_data = b'a' * (2 << 16)
        self.file.manage_upload(file=big_data)
        self.assertIsInstance(str(self.file), str)

    def testViewImageOrFile(self):
        request = self.app.REQUEST
        response = request.RESPONSE
        result = self.file.index_html(request, response)
        self.assertEqual(result, self.data)
        self.assertIsNone(response.getHeader("Content-Disposition"))

    def test_interfaces(self):
        from OFS.Image import Image
        from OFS.interfaces import IWriteLock
        from zope.interface.verify import verifyClass

        verifyClass(IWriteLock, Image)

    def test_text_representation_is_tag(self):
        self.assertEqual(str(self.file),
                         '<img src="http://nohost/file"'
                         ' alt="" title="" height="16" width="16" />')


class SVGTests(ImageTests):
    content_type = 'image/svg+xml'

    def testViewImageOrFile(self):
        request = self.app.REQUEST
        response = request.RESPONSE
        result = self.file.index_html(request, response)
        self.assertEqual(result, self.data)
        self.assertEqual(
            response.getHeader("Content-Disposition"),
            "attachment; filename*=UTF-8''file.svg",
        )

    def testViewImageOrFileNonAscii(self):
        try:
            factory = getattr(self.app, self.factory)
            factory('hällo',
                    file=self.data, content_type=self.content_type)
            transaction.commit()
        except Exception:
            transaction.abort()
            self.connection.close()
            raise
        transaction.begin()
        image = getattr(self.app, 'hällo')
        request = self.app.REQUEST
        response = request.RESPONSE
        result = image.index_html(request, response)
        self.assertEqual(result, self.data)
        self.assertEqual(
            response.getHeader("Content-Disposition"),
            "attachment; filename*=UTF-8''h%C3%A4llo.svg",
        )

    def testViewImageOrFile_with_denylist(self):
        request = self.app.REQUEST
        response = request.RESPONSE
        self.file.use_denylist = True
        result = self.file.index_html(request, response)
        self.assertEqual(result, self.data)
        self.assertEqual(
            response.getHeader("Content-Disposition"),
            "attachment; filename*=UTF-8''file.svg",
        )

    def testViewImageOrFile_with_denylist_and_ct_param(self):
        request = self.app.REQUEST
        response = request.RESPONSE
        self.file.use_denylist = True
        self.file.content_type += ";charset=utf-8"
        result = self.file.index_html(request, response)
        self.assertEqual(result, self.data)
        self.assertEqual(
            response.getHeader("Content-Disposition"),
            "attachment; filename*=UTF-8''file.svg",
        )

    def testViewImageOrFile_with_empty_denylist(self):
        request = self.app.REQUEST
        response = request.RESPONSE
        self.file.use_denylist = True
        self.file.disallowed_inline_mimetypes = []
        result = self.file.index_html(request, response)
        self.assertEqual(result, self.data)
        self.assertIsNone(response.getHeader("Content-Disposition"))

    def test_extract_media_type(self):
        extract = OFS.Image.extract_media_type
        self.assertIsNone(extract(None))
        self.assertEqual(extract("text/plain"), "text/plain")
        self.assertEqual(extract("TEXT/PLAIN"), "text/plain")
        self.assertEqual(extract("text / plain"), "text/plain")
        self.assertEqual(extract(" text/plain ; charset=utf-8"), "text/plain")


class FileEditTests(Testing.ZopeTestCase.FunctionalTestCase):
    """Browser testing ..Image.File"""

    def setUp(self):
        super().setUp()
        uf = self.app.acl_users
        uf.userFolderAddUser('manager', 'manager_pass', ['Manager'], [])
        self.app.manage_addFile('file')

        transaction.commit()
        self.browser = Testing.testbrowser.Browser()
        self.browser.login('manager', 'manager_pass')

    def test_Image__manage_main__1(self):
        """It shows the content of text files as text."""
        self.app.file.update_data('hällo'.encode())
        self.browser.open('http://localhost/file/manage_main')
        text = self.browser.getControl(name='filedata:text').value
        self.assertEqual(text, 'hällo')

    def test_Image__manage_main__3(self):
        """It shows an error message if the file content cannot be decoded."""
        self.app.file.update_data('hällo'.encode('latin-1'))
        self.browser.open('http://localhost/file/manage_main')
        self.assertIn(
            "The file could not be decoded with 'utf-8'.",
            self.browser.contents)

    def test_Image__manage_upload__1(self):
        """It uploads a file, replaces the content and sets content type."""
        self.browser.open('http://localhost/file/manage_main')
        self.browser.getControl(name='file').add_file(
            b'test text file', 'text/plain', 'TestFile.txt')
        self.browser.getControl('Upload File').click()
        self.assertIn('Saved changes', self.browser.contents)
        self.assertEqual(
            self.browser.getControl('Content Type').value, 'text/plain')
        text = self.browser.getControl(name='filedata:text').value
        self.assertEqual(text, 'test text file')

    def test_Image__manage_edit__1(self):
        """It it possible to change the file's content via browser."""
        self.browser.open('http://localhost/file/manage_main')
        text_1 = self.browser.getControl(name='filedata:text').value
        self.assertEqual(text_1, '')
        self.browser.getControl(name='filedata:text').value = 'hällo'
        self.browser.getControl('Save Changes').click()
        self.assertIn('Saved changes', self.browser.contents)
        text_2 = self.browser.getControl(name='filedata:text').value
        self.assertEqual(text_2, 'hällo')

    def test_File__manage_history(self):
        self.app.file.update_data('beföre'.encode())
        transaction.commit()
        self.app.file.update_data('àftér'.encode())
        transaction.commit()
        self.browser.open('http://localhost/file/manage_main')
        self.browser.getLink('History').click()
        keys = self.browser.getControl(name='keys:list')
        keys.value = [keys.options[0], keys.options[1]]
        self.browser.getControl('Compare').click()
        self.assertIn('beföre', self.browser.contents)
        self.assertIn('àftér', self.browser.contents)



"""Testing .subscriber.*"""

import Testing.ZopeTestCase
from Products.SiteAccess.VirtualHostMonster import VirtualHostMonster
from Testing.ZopeTestCase.placeless import temporaryPlacelessSetUp
from Testing.ZopeTestCase.placeless import zcml
from zope.security.management import endInteraction


basic_auth = "{0.user_name}:{0.user_password}".format(Testing.ZopeTestCase)


def setupZCML():
    """Set up the ZCML needed to render the ZMI and the assets."""
    import AccessControl
    import zmi.styles
    import Zope2.App
    zcml.load_config('meta.zcml', Zope2.App)
    zcml.load_config('permissions.zcml', AccessControl)
    zcml.load_config('traversing.zcml', Zope2.App)
    zcml.load_config('configure.zcml', zmi.styles)


class SubscriberTests(Testing.ZopeTestCase.FunctionalTestCase):
    """Test subscribers URL generation with a user from a user folder
    not at the root.
    """
    request_path = f"/{Testing.ZopeTestCase.folder_name}/manage_main"
    resources_base_path = f"/{Testing.ZopeTestCase.folder_name}"

    def afterSetUp(self):
        if "virtual_hosting" not in self.app:
            vhm = VirtualHostMonster()
            vhm.addToContainer(self.app)

    def call_manage_main(self):
        """Call /folder/manage_main and return the HTML text."""

        def _call_manage_main(self):
            self.setRoles(['Manager'])
            # temporaryPlacelessSetUp insists in creating an interaction
            # which the WSGI publisher does not expect.
            endInteraction()
            response = self.publish(self.request_path, basic=basic_auth)
            return str(response)
        return temporaryPlacelessSetUp(
            _call_manage_main, required_zcml=setupZCML)(self)

    def test_subscriber__css_paths__1(self):
        """The paths it returns are rendered in the ZMI."""
        from .subscriber import css_paths
        body = self.call_manage_main()
        for path in css_paths(None):
            self.assertIn(f'href="{self.resources_base_path}{path}"', body)

    def test_subscriber__js_paths__1(self):
        """The paths it returns are rendered in the ZMI."""
        from .subscriber import js_paths
        body = self.call_manage_main()
        for path in js_paths(None):
            self.assertIn(f'src="{self.resources_base_path}{path}"', body)


class SubscriberTestsUserFromRootUserFolderViewingRootFolder(SubscriberTests):
    """Tests subscribers URL generation with a user from the root acl_users,
    viewing the root of ZMI.
    """

    request_path = "/manage_main"
    resources_base_path = ""

    def _setupFolder(self):
        self.folder = self.app

    def _setupUserFolder(self):
        # we use the user folder from self.app
        pass


class SubscriberTestsUserFromRootUserFolderViewingFolder(SubscriberTests):
    """Tests subscribers URL generation with a user from the root acl_users,
    viewing a folder not at the root of ZMI. In such case, the URLs are not
    relative to that folder, the resources are served from the root.
    """

    request_path = f"/{Testing.ZopeTestCase.folder_name}/manage_main"
    resources_base_path = ""

    def _setupUser(self):
        uf = self.app.acl_users
        uf.userFolderAddUser(
            Testing.ZopeTestCase.user_name,
            Testing.ZopeTestCase.user_password,
            ["Manager"],
            [],
        )

    def setRoles(self, roles, name=...):
        # we set roles in _setupUser
        pass

    def login(self):
        pass


class SubscriberUrlWithVirtualHostingTests(SubscriberTests):
    """Tests subscribers URL generation using virtual host."""

    request_path = (
        "/VirtualHostBase/https/example.org:443/VirtualHostRoot/"
        f"{Testing.ZopeTestCase.folder_name}/manage_main"
    )
    resources_base_path = f"/{Testing.ZopeTestCase.folder_name}"


class SubscriberUrlWithVirtualHostingAndUserFolderInVirtualHostTests(
        SubscriberTests):
    """Tests subscribers URL generation using virtual host, when
    the authentication path is part of the virtual host base.
    """

    request_path = (
        "/VirtualHostBase/https/example.org:443/"
        f"{Testing.ZopeTestCase.folder_name}/VirtualHostRoot/manage_main"
    )
    resources_base_path = ""


class SubscriberUrlWithVirtualHostingAndVHTests(SubscriberTests):
    """Tests subscribers URL generation using virtual host, when
    the authentication path is part of the virtual host base and
    using a "_vh_" path element.
    """

    request_path = (
        "/VirtualHostBase/https/example.org:443"
        f"/{Testing.ZopeTestCase.folder_name}/"
        "VirtualHostRoot/_vh_zz/manage_main"
    )
    resources_base_path = "/zz"



import os
from configparser import NoSectionError
from configparser import RawConfigParser


HERE = os.path.abspath(os.path.dirname(__file__))
PYTHON_VERSIONED = {}


class CaseSensitiveParser(RawConfigParser):

    def optionxform(self, value):
        return value


def generate(
        buildout_file: str,
        requirements_file: str,
        zope_requirement: str):
    out_file_requirements = os.path.join(HERE, requirements_file)
    parser = CaseSensitiveParser()
    parser.read(os.path.join(HERE, buildout_file))
    for extend in parser.get('buildout', 'extends', fallback='').splitlines():
        extend = extend.strip()
        if extend:
            parser.read(os.path.join(HERE, extend))

    requirements = []

    # Try to include sections for all currently supported Python versions
    for py_version in ('3.9', '3.10', '3.11', '3.12', '3.13'):
        short_version = py_version.replace('.', '')
        try:
            zope_requirement = _generate(
                parser.items(f'versions:python{short_version}'), py_version,
                requirements, zope_requirement)
        except NoSectionError:
            continue

    # "Unversioned" pins must come last, how they are handled depends on
    # Python version qualifiers for dependencies of the same name.
    zope_requirement = _generate(
        parser.items('versions'), None, requirements, zope_requirement)

    with open(out_file_requirements, 'w') as fd:
        fd.write(zope_requirement)
        for req in sorted(requirements):
            fd.write(req)


def _generate(
    versions: list[tuple[str, str]],
    python_version: str | None,
    requirements: list[str],
    zope_requirement: str
) -> str:
    """Generate requirements or constraints file for a specific Python version.

    If ``python_version`` is false, generate for all python versions.
    Returns a probably changed ``zope_requirement``.
    """
    global PYTHON_VERSIONED  # noqa: F824 name is never assigned in scope

    for name, pin in versions:
        if name == 'Zope':
            if pin:
                zope_requirement = 'Zope==%s\n' % pin
            continue

        if not pin:
            continue

        spec = f'{name}=={pin}'
        if python_version:
            spec = f"{spec}; python_version == '{python_version}'"

            versions = PYTHON_VERSIONED.get(name, set())
            versions.add(python_version)
            PYTHON_VERSIONED[name] = versions
        else:
            if name in PYTHON_VERSIONED:
                versions = sorted(PYTHON_VERSIONED.get(name),
                                  key=lambda s: list(map(int, s.split('.'))))
                spec = f"{spec}; python_version > '{versions[-1]}'"

        requirements.append(spec + '\n')
    return zope_requirement


def main():
    generate(
        'versions-prod.cfg',
        'requirements-full.txt',
        '-e git+https://github.com/zopefoundation/Zope.git@master#egg=Zope\n',
    )
    generate('versions.cfg', 'constraints.txt', '')


if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Converters

Used by `ZPublisher.HTTPRequest` and `OFS.PropertyManager`.

Binary converters (i.e. converters which use `bytes` for/in their result)
are marked with a true `binary` attribute`.
This allows the publisher to perform the conversion to `bytes`
based on its more precise encoding knowledge.
"""


import html
import json
import re
import warnings

from DateTime import DateTime
from DateTime.interfaces import SyntaxError


# This may get overwritten during configuration
default_encoding = 'utf-8'


def field2string(v):
    """Converts value to string."""
    if isinstance(v, bytes):
        return v.decode(default_encoding)
    else:
        return str(v)


def field2bytes(v):
    """Converts value to bytes."""
    if hasattr(v, 'read'):
        return v.read()
    elif isinstance(v, str):
        return v.encode(default_encoding)
    else:
        return bytes(v)


field2bytes.binary = True


def field2text(value, nl=re.compile('\r\n|\n\r').search):
    value = field2string(value)
    match_object = nl(value)
    if match_object is None:
        return value
    length = match_object.start(0)
    result = []
    start = 0
    while length >= start:
        result.append(value[start:length])
        start = length + 2
        match_object = nl(value, start)
        if match_object is None:
            length = -1
        else:
            length = match_object.start(0)

    result.append(value[start:])

    return '\n'.join(result)


def field2required(v):
    v = field2string(v)
    if v.strip():
        return v
    raise ValueError('No input for required field<p>')


def field2int(v):
    if isinstance(v, (list, tuple)):
        return list(map(field2int, v))
    v = field2string(v)
    if v:
        try:
            return int(v)
        except ValueError:
            raise ValueError(
                "An integer was expected in the value %r" % html.escape(
                    v, quote=True
                )
            )
    raise ValueError('Empty entry when <strong>integer</strong> expected')


def field2float(v):
    if isinstance(v, (list, tuple)):
        return list(map(field2float, v))
    v = field2string(v)
    if v:
        try:
            return float(v)
        except ValueError:
            raise ValueError(
                "A floating-point number was expected in the value %r" %
                html.escape(v, True)
            )
    raise ValueError(
        'Empty entry when <strong>floating-point number</strong> expected')


def field2long(v):
    if isinstance(v, (list, tuple)):
        return list(map(field2long, v))
    v = field2string(v)
    # handle trailing 'L' if present.
    if v[-1:] in ('L', 'l'):
        v = v[:-1]
    if v:
        try:
            return int(v)
        except ValueError:
            raise ValueError(
                "A long integer was expected in the value %r" %
                html.escape(
                    v, True))
    raise ValueError('Empty entry when <strong>integer</strong> expected')


def field2tokens(v):
    v = field2string(v)
    return v.split()


def field2lines(v):
    if isinstance(v, (list, tuple)):
        return [field2string(item) for item in v]
    return field2string(v).splitlines()


def field2date(v):
    v = field2string(v)
    try:
        v = DateTime(v)
    except SyntaxError:
        raise SyntaxError("Invalid DateTime " + html.escape(repr(v), True))
    return v


def field2date_international(v):
    v = field2string(v)
    try:
        v = DateTime(v, datefmt="international")
    except SyntaxError:
        raise SyntaxError("Invalid DateTime " + html.escape(repr(v)))
    return v


def field2boolean(v):
    if v == 'False':
        return False
    return bool(v)


def field2ustring(v):
    warnings.warn(
        "The converter `(field2)ustring` is deprecated "
        "and will be removed in Zope 6. "
        "Please use `(field2)string` instead.",
        DeprecationWarning)
    return field2string(v)


def field2utokens(v):
    warnings.warn(
        "The converter `(field2)utokens` is deprecated "
        "and will be removed in Zope 6. "
        "Please use `(field2)tokens` instead.",
        DeprecationWarning)
    return field2tokens(v)


def field2utext(v):
    warnings.warn(
        "The converter `(field2)utext` is deprecated "
        "and will be removed in Zope 6. "
        "Please use `(field2)text` instead.",
        DeprecationWarning)
    return field2text(v)


def field2ulines(v):
    warnings.warn(
        "The converter `(field2u)lines` is deprecated "
        "and will be removed in Zope 6. "
        "Please use `(field2)lines` instead.",
        DeprecationWarning)
    return field2lines(v)


def field2json(v):
    try:
        v = json.loads(v)
    except ValueError:
        raise ValueError("Invalid json " + html.escape(repr(v), True))
    return v


type_converters = {
    'float': field2float,
    'int': field2int,
    'long': field2long,
    'string': field2string,
    'bytes': field2bytes,
    'date': field2date,
    'date_international': field2date_international,
    'json': field2json,
    'required': field2required,
    'tokens': field2tokens,
    'lines': field2lines,
    'text': field2text,
    'boolean': field2boolean,
    'ustring': field2ustring,
    'utokens': field2utokens,
    'ulines': field2ulines,
    'utext': field2utext,
}

get_converter = type_converters.get



##############################################################################
#
# Copyright (c) 2024 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Partial ``ProductContext`` tests.
"""

from types import ModuleType
from types import SimpleNamespace
from unittest import TestCase

from ZPublisher import zpublish_marked

from ..ProductContext import ProductContext


class ProductContextTests(TestCase):
    def setUp(self):
        self.args = _Product(), _App(), ModuleType("pack")
        self.pc = ProductContext(*self.args)
        # ``ProductContext.registerClass`` has a lot of global
        # side effects. We save the original values, set up values
        # for our tests and restore the original values in ``teadDown``
        # When further tests are added, the list likely needs to
        # be extended
        self._saved = {}
        for ospec, val in {
                "sys.modules['Products']": ModuleType("Products"),
                "AccessControl.Permission._registeredPermissions": {},
                "AccessControl.Permission._ac_permissions": (),
                "AccessControl.Permission.ApplicationDefaultPermissions":
                _ApplicationDefaultPermissions}.items():
            obj = _resolve(ospec)
            self._saved[ospec] = obj.get()
            obj.set(val)

    def tearDown(self):
        for ospec, val in self._saved.items():
            _resolve(ospec).set(val)

    def test_initial_tuple(self):

        def c():
            pass

        self.pc.registerClass(meta_type="test", permission="test",
                              constructors=(("name", c),))
        self._verify_reg("name", "c")

    def test_initial_named(self):

        def c():
            pass

        self.pc.registerClass(meta_type="test", permission="test",
                              constructors=(c,))
        self._verify_reg("c", "c")

    def test_constructor_tuple(self):

        def initial():
            pass

        def c():
            pass

        self.pc.registerClass(meta_type="test", permission="test",
                              constructors=(initial, ("name", c),))
        self._verify_reg("name", "c")

    def test_constructor_named(self):

        def initial():
            pass

        def c():
            pass

        self.pc.registerClass(meta_type="test", permission="test",
                              constructors=(initial, c,))
        self._verify_reg("c", "c")

    def test_constructor_noncallable(self):

        def initial():
            pass

        nc = SimpleNamespace(__name__="nc")
        with self.assertWarns(DeprecationWarning):
            self.pc.registerClass(meta_type="test", permission="test",
                                  constructors=(initial, nc))
        self._verify_reg("nc", nc)

    def test_resource_tuple(self):

        def initial():
            pass

        r = SimpleNamespace()
        self.pc.registerClass(meta_type="test", permission="test",
                              constructors=(initial,),
                              resources=(("r", r),))
        self._verify_reg("r", r)

    def test_resource_named(self):

        def initial():
            pass

        r = SimpleNamespace(__name__="r")
        self.pc.registerClass(meta_type="test", permission="test",
                              constructors=(initial,),
                              resources=(r,))
        self._verify_reg("r", r)

    def _verify_reg(self, name, obj):
        pack = self.args[-1]
        m = pack._m
        fo = m[name]
        if isinstance(obj, str):
            self.assertEqual(obj, fo.__name__)
            self.assertTrue(zpublish_marked(fo))
        else:
            self.assertIs(obj, fo)


class _Product:
    """Product mockup."""

    def __init__(self):
        self.id = "pid"


class _App:
    """Application mockup"""


class _ApplicationDefaultPermissions:
    """ApplicationDefaultPermissions mockup."""


class _Attr(SimpleNamespace):
    """an attribute."""

    def get(self):
        return getattr(self.o, self.a)

    def set(self, val):
        setattr(self.o, self.a, val)


class _Item(SimpleNamespace):
    """a (mapping) item."""

    def get(self):
        # we use ``_Item`` for missiong
        return self.o.get(self.a, _Item)

    def set(self, val):
        # we use ``_Item`` for missiong
        if val is _Item:
            del self.o[self.a]
        else:
            self.o[self.a] = val


def _resolve(spec):
    """resolve *spec* into an ``_Attr`` or ``_Item``.

    *spec* is a dotted name, optionally followed by a subscription.
    """
    if "[" in spec:
        dotted, sub = spec[:-1].split("[")
    else:
        dotted, sub = spec, None
    o = None
    segs = dotted.split(".")
    for i in range(0, len(segs) - (0 if sub is not None else 1)):
        seg = segs[i]
        o = getattr(o, seg, None)
        if o is None:
            o = __import__(".".join(segs[:i + 1]), fromlist=(seg,))
    return _Attr(o=o, a=segs[-1]) if sub is None else _Item(o=o, a=eval(sub))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import unittest

from chameleon.exc import ExpressionError

import zope.component.testing
from AccessControl import SecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from Acquisition import Implicit
from Products.PageTemplates.interfaces import IUnicodeEncodingConflictResolver
from Products.PageTemplates.PageTemplate import PageTemplate
from Products.PageTemplates.tests import util
from Products.PageTemplates.unicodeconflictresolver import \
    DefaultUnicodeEncodingConflictResolver
from Products.PageTemplates.unicodeconflictresolver import \
    PreferredCharsetResolver
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from zExceptions import NotFound
from zope.component import provideUtility
from zope.location.interfaces import LocationError
from zope.traversing.adapters import DefaultTraversable

from .util import useChameleonEngine


class AqPageTemplate(Implicit, PageTemplate):
    pass


class AqZopePageTemplate(Implicit, ZopePageTemplate):
    pass


class Folder(util.Base):
    pass


class UnitTestSecurityPolicy:
    """
        Stub out the existing security policy for unit testing purposes.
    """
    # Standard SecurityPolicy interface

    def validate(self,
                 accessed=None,
                 container=None,
                 name=None,
                 value=None,
                 context=None,
                 roles=None,
                 *args, **kw):
        return 1

    def checkPermission(self, permission, object, context):
        return 1


class HTMLTests(zope.component.testing.PlacelessSetup, unittest.TestCase):
    PREFIX = None

    def setUp(self):
        super().setUp()
        useChameleonEngine()
        zope.component.provideAdapter(DefaultTraversable, (None,))

        provideUtility(DefaultUnicodeEncodingConflictResolver,
                       IUnicodeEncodingConflictResolver)

        self.folder = f = Folder()
        f.laf = AqPageTemplate()
        f.t = AqPageTemplate()
        f.z = AqZopePageTemplate('testing')
        self.policy = UnitTestSecurityPolicy()
        self.oldPolicy = SecurityManager.setSecurityPolicy(self.policy)
        noSecurityManager()  # Use the new policy.

    def tearDown(self):
        super().tearDown()
        SecurityManager.setSecurityPolicy(self.oldPolicy)
        noSecurityManager()  # Reset to old policy.

    def assert_expected(self, t, fname, *args, **kwargs):
        t.write(util.read_input(fname))
        assert not t._v_errors, 'Template errors: %s' % t._v_errors
        if self.PREFIX is not None \
                and util.exists_output(self.PREFIX + fname):
            fname = self.PREFIX + fname
        expect = util.read_output(fname)
        out = t(*args, **kwargs)
        util.check_html(expect, out)

    def assert_expected_unicode(self, t, fname, *args, **kwargs):
        t.write(util.read_input(fname))
        assert not t._v_errors, 'Template errors: %s' % t._v_errors
        expect = util.read_output(fname)
        if not isinstance(expect, str):
            expect = str(expect, 'utf-8')
        out = t(*args, **kwargs)
        util.check_html(expect, out)

    def getProducts(self):
        return [
            {'description': 'This is the tee for those who LOVE Zope. '
             'Show your heart on your tee.',
             'price': 12.99, 'image': 'smlatee.jpg'
             },
            {'description': 'This is the tee for Jim Fulton. '
             'He\'s the Zope Pope!',
             'price': 11.99, 'image': 'smpztee.jpg'
             },
        ]

    def test_1(self):
        self.assert_expected(self.folder.laf, 'TeeShopLAF.html')

    def test_2(self):
        self.folder.laf.write(util.read_input('TeeShopLAF.html'))

        self.assert_expected(self.folder.t, 'TeeShop2.html',
                             getProducts=self.getProducts)

    def test_3(self):
        self.folder.laf.write(util.read_input('TeeShopLAF.html'))

        self.assert_expected(self.folder.t, 'TeeShop1.html',
                             getProducts=self.getProducts)

    def testSimpleLoop(self):
        self.assert_expected(self.folder.t, 'Loop1.html')

    def testFancyLoop(self):
        self.assert_expected(self.folder.t, 'Loop2.html')

    def testGlobalsShadowLocals(self):
        self.assert_expected(self.folder.t, 'GlobalsShadowLocals.html')

    def testStringExpressions(self):
        self.assert_expected(self.folder.t, 'StringExpression.html')

    def testReplaceWithNothing(self):
        self.assert_expected(self.folder.t, 'CheckNothing.html')

    def testWithXMLHeader(self):
        self.assert_expected(self.folder.t, 'CheckWithXMLHeader.html')

    def testNotExpression(self):
        self.assert_expected(self.folder.t, 'CheckNotExpression.html')

    def testPathNothing(self):
        self.assert_expected(self.folder.t, 'CheckPathNothing.html')

    def testPathAlt(self):
        self.assert_expected(self.folder.t, 'CheckPathAlt.html')

    def testPathTraverse(self):
        # need to perform this test with a "real" folder
        from OFS.Folder import Folder
        f = self.folder
        self.folder = Folder()
        self.folder.t, self.folder.laf = f.t, f.laf
        self.folder.laf.write('ok')
        self.assert_expected(self.folder.t, 'CheckPathTraverse.html')

    def testBatchIteration(self):
        self.assert_expected(self.folder.t, 'CheckBatchIteration.html')

    def testUnicodeInserts(self):
        self.assert_expected_unicode(self.folder.t, 'CheckUnicodeInserts.html')

    def testI18nTranslate(self):
        self.assert_expected(self.folder.t, 'CheckI18nTranslate.html')

    def testImportOldStyleClass(self):
        self.assert_expected(self.folder.t, 'CheckImportOldStyleClass.html')

    def testRepeatVariable(self):
        self.assert_expected(self.folder.t, 'RepeatVariable.html')

    def testBooleanAttributes(self):
        # Test rendering an attribute that should be empty or left out
        # if the value is non-True
        self.assert_expected(self.folder.t, 'BooleanAttributes.html')

    def testBooleanAttributesAndDefault(self):
        # Zope 2.9 and below support the semantics that an HTML
        # "boolean" attribute (e.g. 'selected', 'disabled', etc.) can
        # be used together with 'default'.
        self.assert_expected(self.folder.t, 'BooleanAttributesAndDefault.html')

    def testInterpolationInContent(self):
        # the chameleon template engine supports ``${path}``
        # interpolations not only as part of ``string`` expressions
        # but globally
        self.assert_expected(self.folder.t, 'InterpolationInContent.html')

    def testBadExpression(self):
        t = self.folder.t
        t.write("<p tal:define='p a//b' />")
        with self.assertRaises(ExpressionError):
            t()

    def testPathAlternativesWithSpaces(self):
        self.assert_expected(self.folder.t, 'PathAlternativesWithSpaces.html')

    def testDefaultKeywordHandling(self):
        self.assert_expected(self.folder.t, 'Default.html')

    def testSwitch(self):
        self.assert_expected(self.folder.t, 'switch.html')

    def test_unicode_conflict_resolution(self):
        # override with the more "demanding" resolver
        provideUtility(PreferredCharsetResolver)
        t = PageTemplate()
        self.assert_expected(t, 'UnicodeResolution.html')

    def test_underscore_traversal(self):
        t = self.folder.t

        t.write('<p tal:define="p context/__class__" />')
        with self.assertRaises(NotFound):
            t()

        # The `random` module used to have something like
        # `import itertools as _itertools`
        # I can't find for which Python versions that was.
        # But in Python 3.9-3.14 the same is true for `tokenize`.
        # But that is not available for direct use in templates.
        # So let's recreate the old situation.
        # Goal is still: test that you can't traverse to this
        # module/attribute that starts with an underscore.
        try:
            from random import _itertools  # NOQA: F401
        except ImportError:
            import itertools
            import random

            random._itertools = itertools

        t.write('<p tal:define="p nocall: random/_itertools/repeat"/>')
        with self.assertRaises((NotFound, LocationError)):
            t()

        t.write('<p tal:content="random/_itertools/repeat/foobar"/>')
        with self.assertRaises((NotFound, LocationError)):
            t()

    def test_module_traversal(self):
        t = self.folder.z

        # Need to reset to the standard security policy so AccessControl
        # checks are actually performed. The test setup initializes
        # a policy that circumvents those checks.
        SecurityManager.setSecurityPolicy(self.oldPolicy)
        noSecurityManager()

        # The getSecurityManager function is explicitly allowed
        t.write(
            '''<p tal:define="a nocall:'''
            '''modules/AccessControl/getSecurityManager"'''
            '''   tal:content="python: a().getUser() or 'Anonymous User'"/>''')
        self.assertEqual(t(), '<p>Anonymous User</p>')

        # Anything else should be unreachable and raise NotFound:
        # Direct access through AccessControl
        t.write('<p tal:define="a nocall:modules/AccessControl/users"/>')
        with self.assertRaises((NotFound, LocationError)):
            t()

        # Indirect access through an intermediary variable
        content = ('<p tal:define="mod nocall:modules/AccessControl;'
                   '               must_fail nocall:mod/users"/>')
        t.write(content)
        with self.assertRaises((NotFound, LocationError)):
            t()

        # Indirect access through an intermediary variable and a dictionary
        content = ('<p tal:define="mod nocall:modules/AccessControl;'
                   '               a_dict python: {\'unsafe\': mod};'
                   '               must_fail nocall: a_dict/unsafe/users"/>')
        t.write(content)
        with self.assertRaises((NotFound, LocationError)):
            t()



import unittest

from zExceptions import Forbidden
from zExceptions import NotFound
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import NotFound as ztkNotFound
from ZPublisher import zpublish
from ZPublisher.BaseRequest import DocstringWarning

from . import docstring


@implementer(IPublishTraverse)
class DummyTraverser:

    def publishTraverse(self, request, name):
        if name == 'dummy':
            return 'dummy object'
        raise ztkNotFound(self, name)


class BaseRequest_factory:

    def _makeOne(self, root):
        from Acquisition import Implicit

        class DummyResponse(Implicit):

            base = ''
            status = None
            debug_mode = False

            def setStatus(self, code):
                self.status = code

            def setBase(self, base):
                if base is None:
                    base = ''
                elif not base.endswith('/'):
                    base = base + '/'
                self.base = str(base)

            def notFoundError(self, name):
                raise NotFound(name)

            forbiddenError = notFoundError

        response = DummyResponse()
        environment = {
            'URL': '',
            'PARENTS': [root],
            'steps': [],
            '_hacked_path': 0,
            '_test_counter': 0,
            'response': response
        }
        return self._getTargetClass()(environment)

    @staticmethod
    def _use_docstring(ud):
        if ud is not None:
            return ud
        from .. import BaseRequest
        return not BaseRequest.deprecate_docstrings

    def _makeBasicObjectClass(self, use_docstring=None):
        from Acquisition import Implicit

        class DummyObjectBasic(Implicit):

            def _setObject(self, id, object):
                setattr(self, id, object)
                return getattr(self, id)

            def view(self):
                # publishable
                return 'view content'

            def noview(self):
                # not publishable
                return 'unpublishable'

            def __contains__(self, name):
                return False

        if self._use_docstring(use_docstring):
            DummyObjectBasic.__doc__ = "with docstring"
            DummyObjectBasic.view.__doc__ = "with docstring"
        else:
            zpublish(DummyObjectBasic)
            zpublish(DummyObjectBasic.view)

        return DummyObjectBasic

    def _makeBasicObject(self, use_docstring=None):
        return self._makeBasicObjectClass(use_docstring)()

    def _makeObjectWithDefault(self, use_docstring=None):

        base = self._makeBasicObjectClass(use_docstring)

        class DummyObjectWithDefault(base):
            def index_html(self):
                # publishable
                return 'index_html content'

        if self._use_docstring(use_docstring):
            DummyObjectWithDefault.__doc__ = "with docstring"
            DummyObjectWithDefault.index_html.__doc__ = "with docstring"
        else:
            zpublish(DummyObjectWithDefault)
            zpublish(DummyObjectWithDefault.index_html)

        return DummyObjectWithDefault()

    def _makeObjectWithDefaultNone(self):

        class DummyObjectWithDefaultNone(self._makeBasicObjectClass()):
            """Dummy class with docstring."""

            index_html = None

        return DummyObjectWithDefaultNone()

    def _makeObjectWithBPTH(self):

        class DummyObjectWithBPTH(self._makeBasicObjectClass()):
            """Dummy class with __before_publishing_traverse__."""

            def __before_publishing_traverse__(self, object, REQUEST):
                if REQUEST['_test_counter'] < 100:
                    REQUEST['_test_counter'] += 1
                else:
                    raise RuntimeError('Infinite loop detected.')
                REQUEST['TraversalRequestNameStack'] += self._path
                REQUEST._hacked_path = 1

        return DummyObjectWithBPTH()

    def _makeObjectWithBD(self):
        class DummyObjectWithBD(self._makeBasicObjectClass()):
            """Dummy class with __browser_default__."""

            def __browser_default__(self, REQUEST):
                if REQUEST['_test_counter'] < 100:
                    REQUEST['_test_counter'] += 1
                else:
                    raise RuntimeError('Infinite loop detected.')
                return self, self._default_path
        return DummyObjectWithBD()

    def _makeObjectWithBBT(self):
        from ZPublisher.interfaces import UseTraversalDefault

        @zpublish
        class _DummyResult:
            ''' '''

            def __init__(self, tag):
                self.tag = tag

        class DummyObjectWithBBT(self._makeBasicObjectClass()):
            """ Dummy class with __bobo_traverse__
            """
            default = _DummyResult('Default')

            def __bobo_traverse__(self, REQUEST, name):
                if name == 'normal':
                    return _DummyResult('Normal')
                elif name == 'default':
                    raise UseTraversalDefault
                raise AttributeError(name)
        return DummyObjectWithBBT()

    def _makeObjectWithBDBBT(self):
        class DummyObjectWithBDBBT(self._makeBasicObjectClass()):
            """Dummy class with __browser_default__."""

            def __browser_default__(self, REQUEST):
                if REQUEST['_test_counter'] < 100:
                    REQUEST['_test_counter'] += 1
                else:
                    raise RuntimeError('Infinite loop detected.')
                return self, self._default_path

            def __bobo_traverse__(self, REQUEST, name):
                if name == self._default_path[0]:
                    return getattr(self, name)
                raise AttributeError(name)
        return DummyObjectWithBDBBT()

    def _makeObjectWithEmptyDocstring(self, use_docstring=None):
        from Acquisition import Implicit

        class DummyObjectWithEmptyDocstring(Implicit):
            def view(self):
                """Attribute with docstring."""
                return 'view content'

            def noview(self):
                # Attribute without docstring.
                return 'unpublishable'
        if self._use_docstring(use_docstring):
            DummyObjectWithEmptyDocstring.__doc__ = ""
        else:
            zpublish(False)(DummyObjectWithEmptyDocstring)
        return DummyObjectWithEmptyDocstring()


class TestBaseRequest(unittest.TestCase, BaseRequest_factory):

    def _getTargetClass(self):
        from ZPublisher.BaseRequest import BaseRequest
        return BaseRequest

    def _makeRootAndFolder(self, use_docstring=None):
        root = self._makeBasicObject(use_docstring)
        folder = root._setObject('folder',
                                 self._makeBasicObject(use_docstring))
        return root, folder

    def test_no_docstring_on_instance(self):
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        self.assertIsNone(r.__doc__)

    def test___bobo_traverse___raises(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objBasic', self._makeBasicObject())
        r = self._makeOne(root)
        self.assertRaises(KeyError, r.__bobo_traverse__, 'REQUEST')
        self.assertRaises(KeyError, r.__bobo_traverse__, 'BODY')
        self.assertRaises(KeyError, r.__bobo_traverse__, 'BODYFILE')

    def test_traverse_basic(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objBasic', self._makeBasicObject())
        r = self._makeOne(root)
        r.traverse('folder/objBasic')
        self.assertEqual(r.URL, '/folder/objBasic')
        self.assertEqual(r.response.base, '')

    def test_traverse_basic_no_acquire_flag(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objBasic', self._makeBasicObject())
        r = self._makeOne(root)
        r['REQUEST_METHOD'] = 'HEAD'
        r.traverse('folder/objBasic')
        self.assertEqual(r.URL, '/folder/objBasic')
        self.assertEqual(r.response.base, '')

    def test_traverse_withDefault(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithDefault', self._makeObjectWithDefault())
        r = self._makeOne(root)
        r.traverse('folder/objWithDefault')
        self.assertEqual(r.URL, '/folder/objWithDefault/index_html')
        self.assertEqual(r.response.base, '/folder/objWithDefault/')

    def test_traverse_withDefaultNone(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithDefaultNone',
                          self._makeObjectWithDefaultNone())
        r = self._makeOne(root)
        r.traverse('folder/objWithDefaultNone')
        self.assertEqual(r.URL, '/folder/objWithDefaultNone')
        self.assertEqual(r.response.base, '')

    def test_traverse_withBPTH(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBPTH', self._makeObjectWithBPTH())
        folder.objWithBPTH._path = ['view']
        r = self._makeOne(root)
        r.traverse('folder/objWithBPTH')
        self.assertEqual(r.URL, '/folder/objWithBPTH/view')
        self.assertEqual(r.response.base, '/folder/objWithBPTH/')

    def test_traverse_withBDView(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBD', self._makeObjectWithBD())
        folder.objWithBD._default_path = ['view']
        r = self._makeOne(root)
        r.traverse('folder/objWithBD')
        self.assertEqual(r.URL, '/folder/objWithBD/view')
        self.assertEqual(r.response.base, '/folder/objWithBD/')

    def test_traverse_withAcquisition(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBPTH', self._makeObjectWithBPTH())
        folder.objWithBPTH._path = ['view']
        folder._setObject('objWithBD', self._makeObjectWithBD())
        folder.objWithBD._default_path = ['view']
        r = self._makeOne(root)
        r.traverse('folder/objWithBD/objWithBPTH')
        self.assertEqual(r.URL, '/folder/objWithBD/objWithBPTH/view')
        self.assertEqual(r.response.base, '/folder/objWithBD/objWithBPTH/')

    def test_traverse_withBDAndBPTH(self):
        # Collector 1079 (infinite loop 1)
        root, folder = self._makeRootAndFolder()
        folder._setObject('objBasic', self._makeBasicObject())
        folder._setObject('objWithBPTH', self._makeObjectWithBPTH())
        folder.objWithBPTH._path = ['objBasic']
        folder._setObject('objWithBD', self._makeObjectWithBD())
        folder.objWithBD._default_path = ['objWithBPTH']
        r = self._makeOne(root)
        r.traverse('folder/objWithBD')
        self.assertEqual(r.URL, '/folder/objWithBD/objWithBPTH/objBasic')
        self.assertEqual(r.response.base, '/folder/objWithBD/objWithBPTH/')

    def test_traverse_withBDEmpty(self):
        # Collector 1079 (infinite loop 2)
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBD', self._makeObjectWithBD())
        folder.objWithBD._default_path = ['']
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objWithBD')

    def test_traverse_withBBT_handles_AttributeError(self):
        # Test that if __bobo_traverse__ raises AttributeError
        # that we get a NotFound
        root, folder = self._makeRootAndFolder()

        def _faux___bobo_traverse__(REQUEST, name):
            raise AttributeError(name)
        obj = self._makeBasicObject()
        obj.__bobo_traverse__ = _faux___bobo_traverse__
        folder._setObject('objWithBBT', obj)
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse,
                          'folder/objWithBBT/bbt_foo')

    def test_traverse_UseTraversalDefault(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBBT', self._makeObjectWithBBT())
        # test non default usage
        r = self._makeOne(root)
        self.assertEqual(r.traverse('folder/objWithBBT/normal').tag, 'Normal')
        # test default usage
        r = self._makeOne(root)
        self.assertEqual(
            r.traverse('folder/objWithBBT/default').tag, 'Default')

    def test_traverse_withBDBBT(self):
        # Test for an object which has a __browser_default__
        # and __bobo_traverse__
        # __bobo_traverse__ should return the object
        # pointed by __browser_default__
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBDBBT', self._makeObjectWithBDBBT())
        folder.objWithBDBBT._default_path = ['view']
        r = self._makeOne(root)
        r.traverse('folder/objWithBDBBT')
        self.assertEqual(r.URL, '/folder/objWithBDBBT/view')
        self.assertEqual(r.response.base, '/folder/objWithBDBBT/')

    def test_traverse_withBDBBT_NotFound(self):
        # Test for an object which has a __browser_default__
        # and __bobo_traverse__
        # __bobo_traverse__ should raise an AttributeError, which will
        # raise a NotFound
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithBDBBT', self._makeObjectWithBDBBT())
        folder.objWithBDBBT._default_path = ['xxx']
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objWithBDBBT')

    def test_traverse_slash(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithDefault', self._makeObjectWithDefault())
        r = self._makeOne(root)
        r['PARENTS'] = [folder.objWithDefault]
        r.traverse('/')
        self.assertEqual(r.URL, '/index_html')
        self.assertEqual(r.response.base, '')

    def test_traverse_special_names(self):
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'REQUEST')
        self.assertRaises(NotFound, r.traverse, 'aq_self')
        self.assertRaises(NotFound, r.traverse, 'aq_base')

    def test_traverse_past_root(self):
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, '..')

    def test_traverse_attribute_with_docstring(self, use_docstring=None):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objBasic', self._makeBasicObject(use_docstring))
        r = self._makeOne(root)
        r.traverse('folder/objBasic/view')
        self.assertEqual(r.URL, '/folder/objBasic/view')
        self.assertEqual(r.response.base, '')

    def test_traverse_attribute_without_docstring(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objBasic', self._makeBasicObject())
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objBasic/noview')

    def test_traverse_acquired_attribute_without_docstring(self):
        root, folder = self._makeRootAndFolder()
        root._setObject('objBasic',
                        self._makeObjectWithEmptyDocstring())
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objBasic')

    def test_traverse_class_without_docstring(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithoutDocstring',
                          self._makeObjectWithEmptyDocstring())
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objWithoutDocstring')

    def test_traverse_attribute_of_class_without_docstring(self):
        root, folder = self._makeRootAndFolder()
        folder._setObject('objWithoutDocstring',
                          self._makeObjectWithEmptyDocstring())
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse,
                          'folder/objWithoutDocstring/view')

    def test_traverse_attribute_and_class_without_docstring(self):
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        folder._setObject('objWithoutDocstring',
                          self._makeObjectWithEmptyDocstring())
        self.assertRaises(NotFound, r.traverse,
                          'folder/objWithoutDocstring/noview')

    def test_traverse_attribute_with_zpublish(self):
        root, folder = self._makeRootAndFolder(False)
        folder._setObject('objBasic', self._makeBasicObject(False))
        r = self._makeOne(root)
        r.traverse('folder/objBasic/view')
        self.assertEqual(r.URL, '/folder/objBasic/view')
        self.assertEqual(r.response.base, '')

    def test_traverse_attribute_without_zpublish(self):
        root, folder = self._makeRootAndFolder(False)
        folder._setObject('objBasic', self._makeBasicObject(False))
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objBasic/noview')

    def test_traverse_acquired_attribute_without_zpublish(self):
        root, folder = self._makeRootAndFolder(False)
        root._setObject('objBasic',
                        self._makeObjectWithEmptyDocstring())
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objBasic')

    def test_traverse_class_without_zpublish(self):
        root, folder = self._makeRootAndFolder(False)
        folder._setObject('objWithoutDocstring',
                          self._makeObjectWithEmptyDocstring(False))
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/objWithoutDocstring')

    def test_traverse_attribute_of_class_without_zpublish(self):
        root, folder = self._makeRootAndFolder(False)
        folder._setObject('objWithoutDocstring',
                          self._makeObjectWithEmptyDocstring(False))
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse,
                          'folder/objWithoutDocstring/view')

    def test_traverse_attribute_and_class_without_zpublish(self):
        root, folder = self._makeRootAndFolder(False)
        r = self._makeOne(root)
        folder._setObject('objWithoutDocstring',
                          self._makeObjectWithEmptyDocstring(False))
        self.assertRaises(NotFound, r.traverse,
                          'folder/objWithoutDocstring/noview')

    def test_docstring_deprecation(self):
        from ZPublisher import BaseRequest
        deprecate = BaseRequest.deprecate_docstrings
        try:
            BaseRequest.deprecate_docstrings = "1"
            with self.assertWarns(BaseRequest.DocstringWarning):
                self.test_traverse_attribute_with_docstring(True)
        finally:
            BaseRequest.deprecate_docstrings = deprecate

    def test_zpublish___call__(self):
        root, folder = self._makeRootAndFolder(False)

        @zpublish(methods="POST")
        def __call__(self):
            pass

        folder.__class__.__call__ = __call__
        # no request method
        r = self._makeOne(root)
        self.assertRaises(Forbidden, r.traverse, 'folder')
        # wrong request method
        r = self._makeOne(root)
        r.environ = dict(REQUEST_METHOD="get")
        self.assertRaises(Forbidden, r.traverse, 'folder')
        # correct request method
        r = self._makeOne(root)
        r.environ = dict(REQUEST_METHOD="post")
        r.traverse('folder')

    def test_traverse_simple_string(self):
        root, folder = self._makeRootAndFolder()
        folder.simpleString = 'foo'
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/simpleString')

    def test_traverse_simple_list(self):
        root, folder = self._makeRootAndFolder()
        folder.simpleList = []
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/simpleList')

    def test_traverse_simple_boolean(self):
        root, folder = self._makeRootAndFolder()
        folder.simpleBoolean = True
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/simpleBoolean')

    def test_traverse_simple_complex(self):
        root, folder = self._makeRootAndFolder()
        folder.simpleComplex = complex(1)
        folder.simpleString = 'foo'
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/simpleComplex')

    def test_traverse_simple_set(self):
        root, folder = self._makeRootAndFolder()
        folder.simpleSet = set()
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/simpleSet')

    def test_traverse_simple_frozen_set(self):
        root, folder = self._makeRootAndFolder()
        folder.simpleFrozenSet = frozenset([])
        r = self._makeOne(root)
        self.assertRaises(NotFound, r.traverse, 'folder/simpleFrozenSet')

    def test_close_w_broken_subscriber(self):
        # See: https://github.com/zopefoundation/Zope/issues/16
        from zope.event import subscribers
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        r.other['foo'] = 'Foo'
        BEFORE = subscribers[:]

        def _broken(event):
            raise ValueError("I'm broken")
        subscribers.append(_broken)
        try:
            self.assertRaises(ValueError, r.close)
        finally:
            subscribers[:] = BEFORE
        self.assertEqual(r.other, {})

    def test_hold_after_close(self):
        # Request should no longer accept holds after it has been closed
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        r._hold(lambda x: None)
        self.assertEqual(len(r._held), 1)
        r.close()
        # No more holding from now on
        self.assertEqual(r._held, None)
        r._hold(lambda x: None)
        self.assertEqual(r._held, None)

    def test_traverse_unsubscriptable(self):
        # See https://bugs.launchpad.net/bugs/213311
        r = self._makeOne(None)
        self.assertRaises(NotFound, r.traverse, 'not_found')

    def test_traverse_publishTraverse(self):
        r = self._makeOne(DummyTraverser())
        self.assertEqual(r.traverse('dummy'), 'dummy object')

    def test_traverse_publishTraverse_error(self):
        r = self._makeOne(DummyTraverser())
        self.assertRaises(NotFound, r.traverse, 'not_found')


class TestRequestViewsBase(unittest.TestCase, BaseRequest_factory):

    _dummy_interface = None

    def _getTargetClass(self):
        from ZPublisher.BaseRequest import BaseRequest
        return BaseRequest

    def _makeOne(self, root):
        from zope.interface import directlyProvides
        from zope.publisher.browser import IDefaultBrowserLayer
        request = super()._makeOne(root)
        # The request needs to implement the proper interface
        directlyProvides(request, IDefaultBrowserLayer)
        return request

    def _makeDummyAclUsers(self):
        from AccessControl.ZopeSecurityPolicy import _noroles
        from Acquisition import Implicit

        class AclUsers(Implicit):
            def validate(self, request, auth='', roles=_noroles):
                # always validate access as anonymous, good for checking
                # if things are publishable regardless of authorization
                from AccessControl.users import nobody
                return nobody.__of__(self)
        acl_users = AclUsers()
        return acl_users

    def _makeRootAndFolder(self):
        root = self._makeBasicObject()
        root.__allow_groups__ = self._makeDummyAclUsers()
        folder = root._setObject('folder', self._makeDummyObject('folder'))
        return root, folder

    def _dummyInterface(self):
        from zope.interface import Interface
        if self._dummy_interface is not None:
            return self._dummy_interface

        class IDummy(Interface):
            """IDummy"""

        self._dummy_interface = IDummy
        return IDummy

    def _makeDummyObject(self, name='dummy'):
        from zope.interface import implementer

        @implementer(self._dummyInterface())
        class DummyObjectZ3(self._makeBasicObjectClass()):

            def __init__(self, name):
                self.name = name

        return DummyObjectZ3(name)

    def _makeDummyObjectWithAttr(self, name):
        from zope.interface import implementer

        @implementer(self._dummyInterface())
        class DummyObjectZ3WithAttr(self._makeBasicObjectClass()):

            def __init__(self, name):
                self.name = name

            def meth(self):
                """doc"""
                return 'meth on %s' % self.name

            def methonly(self):
                """doc"""
                return 'methonly on %s' % self.name

        if not self._use_docstring(None):
            zpublish(DummyObjectZ3WithAttr)
            zpublish(DummyObjectZ3WithAttr.meth)
            zpublish(DummyObjectZ3WithAttr.methonly)

        return DummyObjectZ3WithAttr(name)

    def setUp(self):
        from zope.testing.cleanup import cleanUp

        cleanUp()
        self._registerAdapters()

    def _registerAdapters(self):
        from Acquisition import Implicit
        from zope.component import getGlobalSiteManager
        from zope.interface import Interface
        from zope.publisher.browser import BrowserPage
        from zope.publisher.browser import IDefaultBrowserLayer
        from zope.traversing.interfaces import ITraversable
        from zope.traversing.namespace import view

        gsm = getGlobalSiteManager()

        IDummy = self._dummyInterface()

        class DummyView(Implicit):
            def __init__(self, content, request):
                self.content = content
                self.request = request

            def __call__(self):
                return 'view on %s' % (self.content.name)

        class DummyPage(BrowserPage):
            # BrowserPage is an IBrowserPublisher with a browserDefault that
            # returns self, () so that __call__ is invoked by the publisher.

            def __call__(self):
                return 'Test page'

        class DummyPage2(BrowserPage):

            def browserDefault(self, request):
                # intentionally return something that's not self
                return DummyPage(self.context, request), ()

            # __call__ remains unimplemented,
            # baseclass raises NotImplementedError

        class DummyPage3(BrowserPage):

            def browserDefault(self, request):
                # intentionally return a method here
                return self.foo, ()

            def foo(self):
                return 'Test page'

            # __call__ remains unimplemented,
            # baseclass raises NotImplementedError

        class DummyPage4(Implicit, DummyPage):
            # a normal page that can implicitly acquire attributes
            pass

        # Define the views
        gsm.registerAdapter(DummyView, (IDummy, IDefaultBrowserLayer),
                            Interface, 'meth')
        gsm.registerAdapter(DummyPage, (IDummy, IDefaultBrowserLayer),
                            Interface, 'page')
        gsm.registerAdapter(DummyPage2, (IDummy, IDefaultBrowserLayer),
                            Interface, 'page2')
        gsm.registerAdapter(DummyPage3, (IDummy, IDefaultBrowserLayer),
                            Interface, 'page3')
        gsm.registerAdapter(DummyPage4, (IDummy, IDefaultBrowserLayer),
                            Interface, 'page4')

        # Bind the 'view' namespace (for @@ traversal)
        gsm.registerAdapter(view,
                            (self._dummyInterface(), IDefaultBrowserLayer),
                            ITraversable, 'view')

    def tearDown(self):
        from zope.testing.cleanup import cleanUp
        cleanUp()

    def _setDefaultViewName(self, name):
        from zope.component import getGlobalSiteManager
        from zope.publisher.browser import IBrowserRequest
        from zope.publisher.interfaces import IDefaultViewName
        gsm = getGlobalSiteManager()
        gsm.registerAdapter(name, (self._dummyInterface(), IBrowserRequest),
                            IDefaultViewName, '')


class TestBaseRequestViews(TestRequestViewsBase):

    def test_traverse_view(self):
        # simple view
        root, folder = self._makeRootAndFolder()
        folder._setObject('obj', self._makeDummyObject('obj'))
        r = self._makeOne(root)
        ob = r.traverse('folder/obj/meth')
        self.assertEqual(ob(), 'view on obj')
        ob = r.traverse('folder/obj/@@meth')
        self.assertEqual(ob(), 'view on obj')
        # using default view
        self._setDefaultViewName('meth')
        ob = r.traverse('folder/obj')
        self.assertEqual(ob(), 'view on obj')

    def test_traverse_view_HEAD(self):
        # Make sure traversal to views works for HEAD requests
        root, folder = self._makeRootAndFolder()
        folder._setObject('obj', self._makeDummyObject('obj'))
        r = self._makeOne(root)
        r['REQUEST_METHOD'] = 'HEAD'
        ob = r.traverse('folder/obj/meth')
        self.assertEqual(ob(), 'view on obj')
        ob = r.traverse('folder/obj/@@meth')
        self.assertEqual(ob(), 'view on obj')
        # using default view
        self._setDefaultViewName('meth')
        ob = r.traverse('folder/obj')
        self.assertEqual(ob(), 'view on obj')

    def test_traverse_view_attr_local(self):
        # method on object used first
        root, folder = self._makeRootAndFolder()
        folder._setObject(
            'withattr', self._makeDummyObjectWithAttr('withattr'))
        r = self._makeOne(root)
        ob = r.traverse('folder/withattr/meth')
        self.assertEqual(ob(), 'meth on withattr')
        ob = r.traverse('folder/withattr/@@meth')
        self.assertEqual(ob(), 'view on withattr')
        # using default view
        self._setDefaultViewName('meth')
        ob = r.traverse('folder/withattr')
        self.assertEqual(ob(), 'view on withattr')

    def test_traverse_view_attr_above(self):
        # view takes precedence over acquired attribute
        root, folder = self._makeRootAndFolder()
        folder2 = root._setObject(
            'folder2', self._makeDummyObjectWithAttr('folder2'))
        folder2._setObject('obj2', self._makeDummyObject('obj2'))
        r = self._makeOne(root)
        ob = r.traverse('folder2/obj2/meth')
        self.assertEqual(ob(), 'view on obj2')  # used to be buggy (acquired)
        ob = r.traverse('folder2/obj2/@@meth')
        self.assertEqual(ob(), 'view on obj2')
        # using default view
        self._setDefaultViewName('meth')
        ob = r.traverse('folder2/obj2')
        self.assertEqual(ob(), 'view on obj2')

    def test_traverse_view_attr_local2(self):
        # method with other method above
        root, folder = self._makeRootAndFolder()
        folder2 = root._setObject(
            'folder2', self._makeDummyObjectWithAttr('folder2'))
        folder2._setObject(
            'withattr2', self._makeDummyObjectWithAttr('withattr2'))
        r = self._makeOne(root)
        ob = r.traverse('folder2/withattr2/meth')
        self.assertEqual(ob(), 'meth on withattr2')
        ob = r.traverse('folder2/withattr2/@@meth')
        self.assertEqual(ob(), 'view on withattr2')
        # using default view
        self._setDefaultViewName('meth')
        ob = r.traverse('folder2/withattr2')
        self.assertEqual(ob(), 'view on withattr2')

    def test_traverse_view_attr_acquired(self):
        # normal acquired attribute without view
        root, folder = self._makeRootAndFolder()
        folder2 = root._setObject(
            'folder2', self._makeDummyObjectWithAttr('folder2'))
        folder2._setObject('obj2', self._makeDummyObject('obj2'))
        r = self._makeOne(root)
        ob = r.traverse('folder2/obj2/methonly')
        self.assertEqual(ob(), 'methonly on folder2')
        self.assertRaises(NotFound, r.traverse, 'folder2/obj2/@@methonly')
        # using default view
        self._setDefaultViewName('methonly')
        self.assertRaises(NotFound, r.traverse, 'folder2/obj2')

    def test_quoting_goggles(self):
        # View goggles ('@@') should not be quoted
        root, folder = self._makeRootAndFolder()
        folder._setObject('obj', self._makeDummyObject('obj'))
        r = self._makeOne(root)
        r.traverse('folder/obj/@@meth')
        self.assertEqual(r['URL'], '/folder/obj/@@meth')

    def test_quoting_plusplus(self):
        # View markers ('++ should not be quoted
        root, folder = self._makeRootAndFolder()
        folder._setObject('obj', self._makeDummyObject('obj'))
        r = self._makeOne(root)
        r.traverse('folder/obj/++view++meth')
        self.assertEqual(r['URL'], '/folder/obj/++view++meth')

    def test_browserDefault(self):
        # browserDefault can return self, () to indicate that the
        # object itself wants to be published (using __call__):
        root, folder = self._makeRootAndFolder()
        folder._setObject('obj', self._makeDummyObject('obj'))
        r = self._makeOne(root)
        ob = r.traverse('folder/obj/page')
        self.assertEqual(ob(), 'Test page')

        # browserDefault can return another_object, () to indicate
        # that that object should be published (using __call__):
        r = self._makeOne(root)
        ob = r.traverse('folder/obj/page2')
        self.assertEqual(ob(), 'Test page')

        # browserDefault can also return self.some_method, () to
        # indicate that that method should be called.
        r = self._makeOne(root)
        ob = r.traverse('folder/obj/page3')
        self.assertEqual(ob(), 'Test page')

    def test_wrapping_implicit_acquirers(self):
        # when the default publish traverser finds via adaptation
        # an object providing IAcquirer, it should wrap it in the
        # object being traversed
        root, folder = self._makeRootAndFolder()
        ob2 = self._makeDummyObject('ob2')
        folder._setObject('ob2', ob2)
        r = self._makeOne(root)
        ob = r.traverse('folder/page4')
        self.assertEqual(ob(), 'Test page')
        # make sure we can acquire
        self.assertEqual(ob.ob2, ob2)

    def test__str__returns_native_string(self):
        root, folder = self._makeRootAndFolder()
        r = self._makeOne(root)
        self.assertIsInstance(str(r), str)


class TestDocstringWarning(unittest.TestCase):
    def test_method(self):
        c = docstring.C()
        ds = DocstringWarning(c.g, "URL")
        self.assertEqual(ds.tag(),
                         "'ZPublisher.tests.docstring.C' method "
                         "'f'[ZPublisher.tests.docstring:7]")

    def test_function(self):
        f = docstring.f
        ds = DocstringWarning(f, "URL")
        self.assertEqual(ds.tag(),
                         "function 'ZPublisher.tests.docstring.f' at line 7")

    def test_instance_with_own_docstring(self):
        c = docstring.C()
        c.__doc__ = "c"
        ds = DocstringWarning(c, "URL")
        self.assertEqual(ds.tag(), "object at 'URL'")

    def test_instance_with_inherited_docstring(self):
        c = docstring.C()
        ds = DocstringWarning(c, "URL")
        self.assertEqual(ds.tag(), "'ZPublisher.tests.docstring.C' at line 11")

    def test__str__(self):
        ds = DocstringWarning("special", "URL")
        self.assertEqual(
            str(ds),
            "'builtins.str' uses deprecated docstring "
            "publication control. Use the `ZPublisher.zpublish` decorator "
            "instead")



##############################################################################
#
# Copyright (c) 2023 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" ``PasteDeploy`` filters also known as WSGI middleware.

`The WSGI architecture <https://peps.python.org/pep-3333>`_
consists of a WSGI server, a WSGI application and optionally
WSGI middleware in between.
The server listens for HTTP requests, describes an incoming
request via a WSGI environment, calls the application with
this environment and the function ``start_response`` and sends the
response to the client.
The application is a callable with parameters *environ* and
*start_response*. It processes the request, calls *start_response*
with the response headers and returns an iterable producing the response
body.
A middleware component takes a (base) application and returns an
(enhanced) application.

``PasteDeploy`` calls a middleware component a "filter".
In order to be able to provide parameters, filters are configured
via filter factories. ``paste.deploy`` knows two filter factory types:
``filter_factory`` and ``filter_app_factory``.
A filter_factory has signature ``global_conf, **local_conf`` and
returns a filter (i.e. a function transforming an application into
an application),
a filter_app_factory has signature ``app, global_conf, **local_conf``
and returns the enhanced application directly.
For details see the ``PasteDeploy`` documentation linked from
its PyPI page.

The main content of this module are filter factory definitions.
They are identified by a `filter_` prefix.
Their factory type is determined by the signature.
"""


def filter_content_length(app, global_conf):
    """Honor a ``Content-Length`` header.

    Use this filter if the WSGI server does not follow
    `Note 1 regarding the WSGI input stream
    <https://peps.python.org/pep-3333/#input-and-error-streams>`_
    (such as the ``simple_server`` of Python's ``wsgiref``)
    or violates
    `section 6.3 of RFC 7230
    <https://datatracker.ietf.org/doc/html/rfc7230#section-6.3>`_.
    """
    def enhanced_app(env, start_response):
        wrapped = None
        content_length = env.get("CONTENT_LENGTH")
        if content_length:
            env["wsgi.input"] = wrapped = LimitedFileReader(
                env["wsgi.input"], int(content_length))
        try:
            # Note: this does not special case ``wsgiref.util.FileWrapper``
            # or other similar optimazations
            yield from app(env, start_response)
        finally:
            if wrapped is not None:
                wrapped.discard_remaining()

    return enhanced_app


class LimitedFileReader:
    """File wrapper emulating EOF."""

    # attributes to be delegated to the file
    DELEGATE = {"close", "closed", "fileno", "mode", "name"}

    BUFSIZE = 1 << 14

    def __init__(self, fp, limit):
        """emulate EOF after *limit* bytes have been read.

        *fp* is a binary file like object.
        """
        self.fp = fp
        assert limit >= 0
        self.limit = limit

    def _enforce_limit(self, size):
        limit = self.limit
        return limit if size is None or size < 0 else min(size, limit)

    def read(self, size=-1):
        data = self.fp.read(self._enforce_limit(size))
        self.limit -= len(data)
        return data

    def readline(self, size=-1):
        data = self.fp.readline(self._enforce_limit(size))
        self.limit -= len(data)
        return data

    def __iter__(self):
        return self

    def __next__(self):
        data = self.readline()
        if not data:
            raise StopIteration()
        return data

    def discard_remaining(self):
        while self.read(self.BUFSIZE):
            pass

    def __getattr__(self, attr):
        if attr not in self.DELEGATE:
            raise AttributeError(attr)
        return getattr(self.fp, attr)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Implement an bobo_application object that is BoboPOS3 aware

This module provides a wrapper that causes a database connection to be created
and used when bobo publishes a bobo_application object.
"""


class ZApplicationWrapper:

    def __init__(self, db, name, klass=None):
        self._db = db
        self._name = name
        if klass is not None:
            conn = db.open()
            root = conn.root()
            if name not in root:
                root[name] = klass()
                conn.transaction_manager.commit()
            conn.close()
            self._klass = klass

    # This hack is to overcome a bug in Bobo!
    def __getattr__(self, name):
        return getattr(self._klass, name)

    def __call__(self, connection=None):
        # This is called by ZPublisher.WSGIPublisher,
        # as load_app calls the app object present in the module_info.

        db = self._db
        if connection is None:
            connection = db.open()
        elif isinstance(connection, str):
            connection = db.open(connection)

        return connection.root()[self._name]



import os
import unittest

from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Testing.ZopeTestCase import ZopeTestCase
from zope.component import provideAdapter
from zope.traversing.adapters import DefaultTraversable

from .util import useChameleonEngine


path = os.path.dirname(__file__)


class TestPageTemplateFile(ZopeTestCase):

    def afterSetUp(self):
        useChameleonEngine()
        provideAdapter(DefaultTraversable, (None,))

    def _makeOne(self, name):
        return PageTemplateFile(os.path.join(path, name)).__of__(self.app)

    def test_rr(self):
        class Prioritized:
            __allow_access_to_unprotected_subobjects__ = 1

            def __init__(self, order):
                self.order = order

            def __str__(self):
                return 'P%d' % self.order

        template = self._makeOne('rr.pt')
        result = template(refs=[Prioritized(1), Prioritized(2)])
        self.assertIn('P1', result)
        self.assertLess(result.index('P1'), result.index('P2'))

    def test_locals(self):
        template = self._makeOne('locals.pt')
        result = template()
        self.assertIn('function test', result)
        self.assertIn('function same_type', result)

    def test_locals_base(self):
        template = self._makeOne('locals_base.pt')
        result = template()
        self.assertIn('Application', result)

    def test_nocall(self):
        template = self._makeOne("nocall.pt")

        def dont_call():
            raise RuntimeError()
        result = template(callable=dont_call)
        self.assertIn(repr(dont_call), result)

    def test_exists(self):
        template = self._makeOne("exists.pt")

        def dont_call():
            raise RuntimeError()
        result = template(callable=dont_call)
        self.assertIn('ok', result)

    def test_simple(self):
        template = self._makeOne("simple.pt")
        result = template()
        self.assertIn('Hello world!', result)

    def test_secure(self):
        soup = '<foo></bar>'
        template = self._makeOne("secure.pt")
        from zExceptions import Unauthorized
        try:
            result = template(soup=soup)
        except Unauthorized:
            pass
        else:
            self.fail("Expected unauthorized.")

        from AccessControl.SecurityInfo import allow_module
        allow_module('html')
        result = template(soup=soup)
        self.assertIn('&lt;foo&gt;&lt;/bar&gt;', result)

    def test_structure(self):
        template = self._makeOne("structure.pt")
        param = "<span>abc</span>"
        result = template(param=param)
        self.assertIn(param, result)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(
        TestPageTemplateFile)



##############################################################################
#
# Copyright (c) 2024 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""``zpublish`` related tests."""

from inspect import signature
from unittest import TestCase

from Shared.DC.Scripts.Signature import FuncCode

from .. import zpublish
from .. import zpublish_mark
from .. import zpublish_marked
from .. import zpublish_wrap


class ZpublishTests(TestCase):
    def test_zpublish_true(self):
        @zpublish
        def f():
            pass

        self.assertTrue(zpublish_mark(f))
        self.assertTrue(zpublish_marked(f))

    def test_zpublish_false(self):
        @zpublish(False)
        def f():
            pass

        self.assertFalse(zpublish_mark(f))
        self.assertTrue(zpublish_marked(f))

    def test_zpublish_method(self):
        @zpublish(methods="method")
        def f():
            pass

        self.assertEqual(zpublish_mark(f), ("METHOD",))
        self.assertTrue(zpublish_marked(f))

    def test_zpublish_methods(self):
        @zpublish(methods="m1 m2".split())
        def f():
            pass

        self.assertEqual(zpublish_mark(f), ("M1", "M2"))
        self.assertTrue(zpublish_marked(f))

    def test_zpublish_mark(self):
        def f():
            pass

        self.assertIsNone(zpublish_mark(f))
        self.assertTrue(zpublish_mark(f, True))
        zpublish(f)
        self.assertTrue(zpublish_mark(f))

    def test_zpublish_marked(self):
        def f():
            pass

        self.assertFalse(zpublish_marked(f))
        zpublish(f)
        self.assertTrue(zpublish_marked(f))

    def test_zpublish_wrap(self):
        def f():
            pass

        self.assertFalse(zpublish_marked(f))
        wrapper = zpublish_wrap(f)
        self.assertFalse(zpublish_marked(f))
        self.assertTrue(zpublish_mark(wrapper))
        self.assertEqual(signature(wrapper), signature(f))
        self.assertIs(wrapper, zpublish_wrap(wrapper))
        wrapper2 = zpublish_wrap(wrapper, conditional=False, methods="put")
        self.assertIsNot(wrapper2, wrapper)
        self.assertEqual(zpublish_mark(wrapper2), ("PUT",))

        # test ``mapply`` signature

        class WithMapplySignature:
            __code__ = FuncCode(("a", "b", "c"), 2)
            __defaults__ = None

            def __call__(self, *args, **kw):
                pass

        f = WithMapplySignature()
        wrapper = zpublish_wrap(f)
        self.assertEqual(str(wrapper.__signature__), "(a, b)")
        WithMapplySignature.__defaults__ = 2,
        wrapper = zpublish_wrap(f)
        self.assertEqual(str(wrapper.__signature__), "(a, b=2)")



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
""" Some utility functions for unit tests
"""
import contextlib
import sys


@contextlib.contextmanager
def capture_stdout(file):
    old_out = sys.stdout
    sys.stdout = file
    yield
    sys.stdout = old_out



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Component tests
"""

import unittest
from doctest import DocFileSuite

from Testing.ZopeTestCase import FunctionalDocFileSuite


def test_suite():
    return unittest.TestSuite([
        DocFileSuite('component.txt', package="Products.Five.component"),
        FunctionalDocFileSuite('makesite.txt',
                               package="Products.Five.component"),
    ])



# (c) 2005 Ian Bicking and contributors; written for Paste
# (http://pythonpaste.org) Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php
#
# For discussion of daemonizing:
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/278731
#
# Code taken also from QP: http://www.mems-exchange.org/software/qp/ From
# lib/site.py
#
# Code further taken from:
# https://github.com/Pylons/pyramid/blob/master/pyramid/scripts/pserve.py and
# https://github.com/Pylons/pyramid/blob/master/pyramid/paster.py
# licensed under the BSD-derived Repoze Public License
# (http://repoze.org/license.html).

import configparser
import optparse
import os
import re
import sys
from logging.config import fileConfig

from paste.deploy import loadapp
from paste.deploy import loadserver

import Zope2
from App.config import getConfiguration


def parse_vars(args):
    """
    Given variables like ``['a=b', 'c=d']`` turns it into ``{'a':
    'b', 'c': 'd'}``
    """
    result = {}
    for arg in args:
        if '=' not in arg:
            raise ValueError(
                'Variable assignment %r invalid (no "=")'
                % arg)
        name, value = arg.split('=', 1)
        result[name] = value
    return result


def _getpathsec(config_uri, name):
    if '#' in config_uri:
        path, section = config_uri.split('#', 1)
    else:
        path, section = config_uri, 'main'
    if name:
        section = name
    return path, section


def setup_logging(config_uri, global_conf=None,  # NOQA
                  fileConfig=fileConfig,
                  configparser=configparser):
    """
    Set up logging via :func:`logging.config.fileConfig` with the filename
    specified via ``config_uri`` (a string in the form
    ``filename#sectionname``).
    ConfigParser defaults are specified for the special ``__file__``
    and ``here`` variables, similar to PasteDeploy config loading.
    Extra defaults can optionally be specified as a dict in ``global_conf``.
    """
    path, _ = _getpathsec(config_uri, None)
    parser = configparser.ConfigParser()
    parser.read([path])
    if parser.has_section('loggers'):
        config_file = os.path.abspath(path)
        full_global_conf = dict(
            __file__=config_file,
            here=os.path.dirname(config_file))
        if global_conf:
            full_global_conf.update(global_conf)
        return fileConfig(
            config_file, full_global_conf, disable_existing_loggers=False)


class ServeCommand:

    usage = '%prog config_uri [var=value]'
    description = """\
This command serves a web application that uses a PasteDeploy
configuration file for the server and application.
You can also include variable assignments like 'http_port=8080'
and then use %(http_port)s in your config files.
    """
    default_verbosity = 1

    parser = optparse.OptionParser(
        usage,
        description=description
    )
    parser.add_option(
        '-n', '--app-name',
        dest='app_name',
        metavar='NAME',
        help="Load the named application (default main)")
    parser.add_option(
        '-s', '--server',
        dest='server',
        metavar='SERVER_TYPE',
        help="Use the named server.")
    parser.add_option(
        '--server-name',
        dest='server_name',
        metavar='SECTION_NAME',
        help=("Use the named server as defined in the configuration file "
              "(default: main)"))
    parser.add_option(
        '-v', '--verbose',
        default=default_verbosity,
        dest='verbose',
        action='count',
        help="Set verbose level (default " + str(default_verbosity) + ")")
    parser.add_option(
        '-q', '--quiet',
        action='store_const',
        const=0,
        dest='verbose',
        help="Suppress verbose output")
    parser.add_option(
        '-d', '--debug',
        action='store_const',
        const=1,
        dest='debug',
        help="Enable debug mode.")
    parser.add_option(
        '-e', '--debug-exceptions',
        action='store_const',
        const=1,
        dest='debug_exceptions',
        help="Enable exceptions debug mode.")

    _scheme_re = re.compile(r'^[a-z][a-z]+:', re.I)

    def __init__(self, argv, quiet=False):
        self.options, self.args = self.parser.parse_args(argv[1:])
        if quiet:
            self.options.verbose = 0

    def out(self, msg):
        if self.options.verbose > 0:
            print(msg)

    def get_options(self):
        restvars = self.args[1:]
        return parse_vars(restvars)

    def run(self):
        if not self.args:
            self.out('You must give a config file')
            return 2
        app_spec = self.args[0]
        app_name = self.options.app_name

        vars = self.get_options()

        if not self._scheme_re.search(app_spec):
            app_spec = 'config:' + app_spec
        server_name = self.options.server_name
        if self.options.server:
            server_spec = 'egg:Zope2'
            assert server_name is None
            server_name = self.options.server
        else:
            server_spec = app_spec
        base = os.getcwd()

        log_fn = app_spec
        if log_fn.startswith('config:'):
            log_fn = app_spec[len('config:'):]
        elif log_fn.startswith('egg:'):
            log_fn = None
        if log_fn:
            log_fn = os.path.join(base, log_fn)
            setup_logging(log_fn, global_conf=vars)

        server = self.loadserver(server_spec, name=server_name,
                                 relative_to=base, global_conf=vars)

        if 'debug_mode' not in vars and self.options.debug:
            vars['debug_mode'] = 'true'
        if 'debug_exceptions' not in vars and self.options.debug_exceptions:
            vars['debug_exceptions'] = 'true'
        app = self.loadapp(app_spec, name=app_name, relative_to=base,
                           global_conf=vars)

        if self.options.verbose > 0:
            if hasattr(os, 'getpid'):
                msg = 'Starting server in PID %i.' % os.getpid()
            else:
                msg = 'Starting server.'
            self.out(msg)

        def serve():
            self.makePidFile()

            try:
                server(app)
            except (SystemExit, KeyboardInterrupt) as e:
                if self.options.verbose > 1:
                    raise
                if str(e):
                    msg = ' ' + str(e)
                else:
                    msg = ''
                self.out('Exiting%s (-v to see traceback)' % msg)
            finally:
                for db in Zope2.opened:
                    db.close()
                self.unlinkPidFile()

        serve()

    def loadapp(self, app_spec, name, relative_to, **kw):
        return loadapp(app_spec, name=name, relative_to=relative_to, **kw)

    def loadserver(self, server_spec, name, relative_to, **kw):
        return loadserver(
            server_spec, name=name, relative_to=relative_to, **kw)

    def makePidFile(self):
        options = getConfiguration()
        try:
            IO_ERRORS = (IOError, OSError, WindowsError)
        except NameError:
            IO_ERRORS = (IOError, OSError)

        try:
            if os.path.exists(options.pid_filename):
                os.unlink(options.pid_filename)
            with open(options.pid_filename, 'w') as fp:
                fp.write(str(os.getpid()))
        except IO_ERRORS:
            pass

    def unlinkPidFile(self):
        options = getConfiguration()
        try:
            os.unlink(options.pid_filename)
        except OSError:
            pass


def main(argv=sys.argv, quiet=False):
    command = ServeCommand(argv, quiet=quiet)
    return command.run()


if __name__ == '__main__':
    sys.exit(main() or 0)



##############################################################################
#
# Copyright (c) 2023 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest
from io import BytesIO

from paste.deploy import loadfilter

from ..pastefilter import LimitedFileReader


class TestLimitedFileReader(unittest.TestCase):
    def test_enforce_limit(self):
        f = LimitedFileReader(BytesIO(), 10)
        enforce = f._enforce_limit
        self.assertEqual(enforce(None), 10)
        self.assertEqual(enforce(-1), 10)
        self.assertEqual(enforce(20), 10)
        self.assertEqual(enforce(5), 5)

    def test_read(self):
        f = LimitedFileReader(BytesIO(b"123\n567\n901\n"), 10)
        self.assertEqual(len(f.read()), 10)
        self.assertEqual(len(f.read()), 0)
        f = LimitedFileReader(BytesIO(b"123\n567\n901\n"), 10)
        self.assertEqual(len(f.read(8)), 8)
        self.assertEqual(len(f.read(3)), 2)
        self.assertEqual(len(f.read(3)), 0)

    def test_readline(self):
        f = LimitedFileReader(BytesIO(b"123\n567\n901\n"), 10)
        self.assertEqual(f.readline(), b"123\n")
        self.assertEqual(f.readline(), b"567\n")
        self.assertEqual(f.readline(), b"90")
        self.assertEqual(f.readline(), b"")
        f = LimitedFileReader(BytesIO(b"123\n567\n901\n"), 10)
        self.assertEqual(f.readline(1), b"1")

    def test_iteration(self):
        f = LimitedFileReader(BytesIO(b"123\n567\n901\n"), 10)
        self.assertEqual(list(f), [b"123\n", b"567\n", b"90"])

    def test_discard_remaining(self):
        fp = BytesIO(b"123\n567\n901\n")
        LimitedFileReader(fp, 10).discard_remaining()
        self.assertEqual(fp.read(), b"1\n")

    def test_delegation(self):
        f = LimitedFileReader(BytesIO(b"123\n567\n901\n"), 10)
        with self.assertRaises(AttributeError):
            f.write
        f.close()


class TestFilters(unittest.TestCase):
    def test_content_length(self):
        filter = loadfilter("egg:Zope", "content_length")

        def app(env, start_response):
            return iter((env["wsgi.input"],))

        def request(env, app=filter(app)):
            return app(env, None)

        fp = BytesIO()
        env = {"wsgi.input": fp}
        self.assertIs(next(request(env)), fp)

        fp = BytesIO(b"123")
        env = {"wsgi.input": fp}
        env["CONTENT_LENGTH"] = "3"
        response = request(env)
        r = next(response)
        self.assertIsInstance(r, LimitedFileReader)
        self.assertEqual(r.limit, 3)
        with self.assertRaises(StopIteration):
            next(response)
        self.assertFalse(fp.read())



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import base64
import logging

import transaction
from Acquisition import aq_inner
from Acquisition import aq_parent


logger = logging.getLogger('ZPublisher')
AC_LOGGER = logging.getLogger('event.AccessControl')


def recordMetaData(object, request):
    if hasattr(object, 'getPhysicalPath'):
        path = '/'.join(object.getPhysicalPath())
    else:
        # Try hard to get the physical path of the object,
        # but there are many circumstances where that's not possible.
        to_append = ()

        if hasattr(object, '__self__') and hasattr(object, '__name__'):
            # object is a Python method.
            to_append = (object.__name__,)
            object = object.__self__

        while object is not None and not hasattr(object, 'getPhysicalPath'):
            if getattr(object, '__name__', None) is None:
                object = None
                break
            to_append = (object.__name__,) + to_append
            object = aq_parent(aq_inner(object))

        if object is not None:
            path = '/'.join(object.getPhysicalPath() + to_append)
        else:
            # As Jim would say, "Waaaaaaaa!"
            # This may cause problems with virtual hosts
            # since the physical path is different from the path
            # used to retrieve the object.
            path = request.get('PATH_INFO')

    T = transaction.get()
    T.note(safe_unicode(path))
    auth_user = request.get('AUTHENTICATED_USER', None)
    if auth_user:
        auth_folder = aq_parent(auth_user)
        if auth_folder is None:
            AC_LOGGER.warning(
                'A user object of type %s has no aq_parent.',
                type(auth_user))
            auth_path = request.get('AUTHENTICATION_PATH')
        else:
            auth_path = '/'.join(auth_folder.getPhysicalPath()[1:-1])
        user_id = auth_user.getId()
        user_id = safe_unicode(user_id) if user_id else 'None'
        T.setUser(user_id, safe_unicode(auth_path))


def safe_unicode(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, bytes):
        try:
            value = str(value, 'utf-8')
        except UnicodeDecodeError:
            value = value.decode('utf-8', 'replace')
    return value


def basic_auth_encode(user, password=None):
    # user / password and the return value are of type str
    value = user
    if password is not None:
        value = value + ':' + password
    header = b'Basic ' + base64.b64encode(value.encode())
    header = header.decode()
    return header


def basic_auth_decode(token):
    # token and the return values are of type str
    if not token:
        return None
    if not token[:6].lower() == 'basic ':
        return None
    value = token.split()[-1]  # Strip 'Basic '
    plain = base64.b64decode(value).decode()
    user, password = plain.split(':', 1)  # Split at most once
    return (user, password)


def _string_tuple(value):
    if not value:
        return ()
    return tuple([safe_unicode(element) for element in value])


def fix_properties(obj, path=None):
    """Fix properties on object.

    This does two things:

    1. Make sure lines contain only strings, instead of bytes,
       or worse: a combination of strings and bytes.
    2. Replace deprecated ulines, utext, utoken, and ustring properties
       with their non-unicode variant, using native strings.

    See https://github.com/zopefoundation/Zope/issues/987

    Since Zope 5.3, a lines property stores strings instead of bytes.
    But there is no migration yet.  (We do that here.)
    Result is that getProperty on an already created lines property
    will return the old value with bytes, but a newly created lines property
    will return strings.  And you might get combinations.

    Also since Zope 5.3, the ulines property type is deprecated.
    You should use a lines property instead.
    Same for a few others: utext, utoken, ustring.
    The unicode variants are planned to be removed in Zope 6.

    Intended usage:
    app.ZopeFindAndApply(app, apply_func=fix_properties)
    """
    if path is None:
        # When using ZopeFindAndApply, path is always given.
        # But we may be called by other code.
        if hasattr(object, 'getPhysicalPath'):
            path = '/'.join(object.getPhysicalPath())
        else:
            # Some simple object, for example in tests.
            # We don't care about the path then, it is only shown in logs.
            path = "/dummy"

    if not hasattr(obj, "_updateProperty"):
        # Seen with portal_url tool, most items in portal_skins,
        # catalog lexicons, workflow states/transitions/variables, etc.
        return
    try:
        prop_map = obj.propertyMap()
    except (AttributeError, TypeError, KeyError, ValueError):
        # If getting the property map fails, there is nothing we can do.
        # Problems seen in practice:
        # - Object does not inherit from PropertyManager,
        #   for example 'MountedObject'.
        # - Object is a no longer existing skin layer.
        logger.warning("Error getting property map from %s", path)
        return

    for prop_info in prop_map:
        # Example: {'id': 'title', 'type': 'string', 'mode': 'w'}
        prop_id = prop_info.get("id")
        current = obj.getProperty(prop_id)
        if current is None:
            continue
        new_type = prop_type = prop_info.get("type")
        if prop_type == "lines":
            new = _string_tuple(current)
        elif prop_type == "ulines":
            new_type = "lines"
            new = _string_tuple(current)
        elif prop_type == "utokens":
            new_type = "tokens"
            new = _string_tuple(current)
        elif prop_type == "utext":
            new_type = "text"
            new = safe_unicode(current)
        elif prop_type == "ustring":
            new_type = "string"
            new = safe_unicode(current)
        else:
            continue
        if prop_type != new_type:
            # Replace with non-unicode variant.
            # This could easily lead to:
            # Exceptions.BadRequest: Invalid or duplicate property id.
            #   obj._delProperty(prop_id)
            #   obj._setProperty(prop_id, new, new_type)
            # So fix it by using internal details.
            for prop in obj._properties:
                if prop.get("id") == prop_id:
                    prop["type"] = new_type
                    obj._p_changed = True
                    break
            else:
                # This probably cannot happen.
                # If it does, we want to know.
                logger.warning(
                    "Could not change property %s from %s to %s for %s",
                    prop_id,
                    prop_type,
                    new_type,
                    path,
                )
                continue
            obj._updateProperty(prop_id, new)
            logger.info(
                "Changed property %s from %s to %s for %s",
                prop_id,
                prop_type,
                new_type,
                path,
            )
            continue
        if current != new:
            obj._updateProperty(prop_id, new)
            logger.info(
                "Changed property %s at %s so value fits the type %s: %r",
                prop_id,
                path,
                prop_type,
                new,
            )



import itertools

import zope.component
import zope.interface
from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_parent


def prepend_authentication_path(context, path):
    """Prepend the path of the user folder.

    Because ++resource++zmi is protected, we generate URLs relative to the
    user folder of the logged-in user, so that the user can access the
    resources.
    """
    request = getattr(context, 'REQUEST', None)
    if request is None:
        return path
    user_folder = aq_parent(getSecurityManager().getUser())
    if user_folder is None:
        return path
    # prepend the authentication path, unless it is already part of the
    # virtual host root.
    authentication_path = []
    ufpp = aq_parent(user_folder).getPhysicalPath()
    vrpp = request.get("VirtualRootPhysicalPath") or ()
    for ufp, vrp in itertools.zip_longest(ufpp, vrpp):
        if ufp == vrp:
            continue
        authentication_path.append(ufp)

    parts = [
        p for p in itertools.chain(authentication_path, path.split("/")) if p]

    return request.physicalPathToURL(parts, relative=True)


@zope.component.adapter(zope.interface.Interface)
def css_paths(context):
    """Return paths to CSS files needed for the Zope 4 ZMI."""
    return (
        prepend_authentication_path(context, path)
        for path in (
            '/++resource++zmi/bootstrap-4.6.0/bootstrap.min.css',
            '/++resource++zmi/fontawesome-free-5.15.2/css/all.css',
            '/++resource++zmi/zmi_base.css',
        )
    )


@zope.component.adapter(zope.interface.Interface)
def js_paths(context):
    """Return paths to JS files needed for the Zope 4 ZMI."""
    return (
        prepend_authentication_path(context, path)
        for path in (
            '/++resource++zmi/jquery-3.5.1.min.js',
            '/++resource++zmi/bootstrap-4.6.0/bootstrap.bundle.min.js',
            '/++resource++zmi/ace.ajax.org/ace.js',
            '/++resource++zmi/zmi_base.js',
            '/++resource++zmi/zmi.localstorage.api.js',
        )
    )



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import os
import os.path
import re
import sys
import unittest

from ExtensionClass import Base
from Products.PageTemplates.engine import Program
from zope.component import provideUtility
from zope.pagetemplate.interfaces import IPageTemplateEngine
from zope.pagetemplate.pagetemplate import PageTemplateEngine


# Dummy TestCase to use the assertions outside the actual tests.
TEST_CASE = unittest.TestCase('__init__')


class Bruce(Base):
    __allow_access_to_unprotected_subobjects__ = 1
    isDocTemp = 0

    def __str__(self):
        return 'bruce'

    def __int__(self):
        return 42

    def __float__(self):
        return 42.0

    def keys(self):
        return ['bruce'] * 7

    def values(self):
        return [self] * 7

    def items(self):
        return [('bruce', self)] * 7

    def __len__(self):
        return 7

    def __getitem__(self, index):
        if isinstance(index, int) and (index < 0 or index > 6):
            raise IndexError(index)
        return self

    def __getattr__(self, name):
        if name[:1] == '_':
            raise AttributeError(name)
        return self


bruce = Bruce()


class arg(Base):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, nn, aa):
        self.num, self.arg = nn, aa

    def __str__(self):
        return str(self.arg)


class argv(Base):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, argv=sys.argv[1:]):
        args = self.args = []
        for aa in argv:
            args.append(arg(len(args) + 1, aa))

    def items(self):
        return [('spam%d' % a.num, a) for a in self.args]

    def values(self):
        return self.args

    def getPhysicalRoot(self):
        return self


def check_html(s1, s2):
    if not isinstance(s2, bytes) and isinstance(s1, bytes):
        # convert to common type
        s1 = s1.decode("utf-8")  # our encoding
    s1 = normalize_html(s1)
    s2 = normalize_html(s2)
    TEST_CASE.assertEqual(s1, s2)


def check_xml(s1, s2):
    s1 = normalize_xml(s1)
    s2 = normalize_xml(s2)
    TEST_CASE.assertEqual(s1, s2, "XML Output Changed")


def normalize_html(s):
    s = re.sub(r"[ \t]+", " ", s)
    s = re.sub(r"/>", ">", s)
    return s


def normalize_xml(s):
    s = re.sub(r"\s+", " ", s)
    s = re.sub(r"(?s)\s+<", "<", s)
    s = re.sub(r"(?s)>\s+", ">", s)
    return s


HERE = os.path.dirname(__file__)
input_dir = os.path.join(HERE, 'input')
output_dir = os.path.join(HERE, 'output')


def _open(filename, mode):
    # Define explicit encoding for windows platform
    return open(filename, mode, encoding='utf-8')


def read_input(filename):
    filename = os.path.join(input_dir, filename)
    with _open(filename, 'r') as fd:
        data = fd.read()
    return data


def read_output(filename):
    filename = os.path.join(output_dir, filename)
    with _open(filename, 'r') as fd:
        data = fd.read()
    return data


def exists_output(filename):
    filename = os.path.join(output_dir, filename)
    return os.path.exists(filename)


def useChameleonEngine():
    # Force the use of the new chameleon rendering engine (the new default).
    # Its use depends on a utility registration that is queried in
    # zope.pagetemplate,pagetemplate.PageTemplate's _cook method. Unfortunately
    # the fallback is the old Zope engine if there is no registration, so we
    # force one here for use by unit tests.
    provideUtility(Program, IPageTemplateEngine)


def useOldZopeEngine():
    # BBB Force the use of the old Zope page template engine, which is needed
    # for some tests that test features only supported by it.
    provideUtility(PageTemplateEngine, IPageTemplateEngine)



##############################################################################
#
# Copyright (c) 2017 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest


class SafeUnicodeTests(unittest.TestCase):

    def _makeOne(self, value):
        from ZPublisher.utils import safe_unicode
        return safe_unicode(value)

    def test_ascii(self):
        self.assertEqual(self._makeOne('foo'), 'foo')
        self.assertEqual(self._makeOne(b'foo'), 'foo')

    def test_latin_1(self):
        self.assertEqual(self._makeOne(b'fo\xf6'), 'fo\ufffd')

    def test_unicode(self):
        self.assertEqual(self._makeOne('foö'), 'foö')

    def test_utf_8(self):
        self.assertEqual(self._makeOne('test\xc2\xae'), 'test\xc2\xae')
        self.assertEqual(self._makeOne(b'test\xc2\xae'), 'test\xae')


class NoUpdatePropertyManager:
    """PropertyManager without _updateProperty method.

    This is a simplified version of the original PropertyManager,
    with only the methods we need.
    """
    _properties = ()

    def _setPropValue(self, id, value):
        if isinstance(value, list):
            value = tuple(value)
        setattr(self, id, value)

    def _setProperty(self, id, value, type='string'):
        self._properties = self._properties + ({'id': id, 'type': type},)
        self._setPropValue(id, value)

    def hasProperty(self, id):
        for p in self._properties:
            if id == p['id']:
                return 1
        return 0

    def getProperty(self, id, d=None):
        if self.hasProperty(id):
            return getattr(self, id)
        return d

    def getPropertyType(self, id):
        for md in self._properties:
            if md['id'] == id:
                return md.get('type', 'string')
        return None

    def _propertyMap(self):
        return self._properties

    def propertyMap(self):
        return tuple(dict.copy() for dict in self._propertyMap())


class NoPropertiesManager(NoUpdatePropertyManager):
    """PropertyManager with _updateProperty method but without _properties."""
    _properties = None

    def _updateProperty(self, id, value):
        self._setPropValue(id, value)


class FixPropertiesTests(unittest.TestCase):

    def _makeOne(self):
        from OFS.PropertyManager import PropertyManager

        return PropertyManager()

    def test_lines(self):
        from ZPublisher.utils import fix_properties

        obj = self._makeOne()
        obj._setProperty("mixed", ["text and", b"bytes"], "lines")
        self.assertEqual(obj.getProperty("mixed"), ("text and", b"bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "lines")

        fix_properties(obj)
        self.assertEqual(obj.getProperty("mixed"), ("text and", "bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "lines")

    def test_ulines(self):
        from ZPublisher.utils import fix_properties

        obj = self._makeOne()
        obj._setProperty("mixed", ["text and", b"bytes"], "ulines")
        self.assertEqual(obj.getProperty("mixed"), ("text and", b"bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "ulines")

        fix_properties(obj)
        self.assertEqual(obj.getProperty("mixed"), ("text and", "bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "lines")

    def test_utokens(self):
        from ZPublisher.utils import fix_properties

        obj = self._makeOne()
        obj._setProperty("mixed", ["text", "and", b"bytes"], "utokens")
        self.assertEqual(obj.getProperty("mixed"), ("text", "and", b"bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "utokens")

        fix_properties(obj)
        self.assertEqual(obj.getProperty("mixed"), ("text", "and", "bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "tokens")

    def test_utext(self):
        from ZPublisher.utils import fix_properties

        obj = self._makeOne()
        obj._setProperty("prop1", "multiple\nlines", "utext")
        self.assertEqual(obj.getProperty("prop1"), "multiple\nlines")
        self.assertEqual(obj.getPropertyType("prop1"), "utext")

        fix_properties(obj)
        self.assertEqual(obj.getProperty("prop1"), "multiple\nlines")
        self.assertEqual(obj.getPropertyType("prop1"), "text")

    def test_ustring(self):
        from ZPublisher.utils import fix_properties

        obj = self._makeOne()
        obj._setProperty("prop1", "single line", "ustring")
        self.assertEqual(obj.getProperty("prop1"), "single line")
        self.assertEqual(obj.getPropertyType("prop1"), "ustring")

        fix_properties(obj)
        self.assertEqual(obj.getProperty("prop1"), "single line")
        self.assertEqual(obj.getPropertyType("prop1"), "string")

    def test_no_update(self):
        # Test that an object without _updateProperty method does not trip up
        # our code.
        from ZPublisher.utils import fix_properties

        obj = NoUpdatePropertyManager()
        obj._setProperty("mixed", ["text and", b"bytes"], "lines")
        self.assertEqual(obj.getProperty("mixed"), ("text and", b"bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "lines")

        # This should not raise an error.
        fix_properties(obj)
        # The properties should have remained the same.
        self.assertEqual(obj.getProperty("mixed"), ("text and", b"bytes"))
        self.assertEqual(obj.getPropertyType("mixed"), "lines")

    def test_no_properties(self):
        # Test that an object with a failing propertyMap method,
        # due to _properties=None, does not trip up our code.
        from ZPublisher.utils import fix_properties

        obj = NoPropertiesManager()
        # This should not raise an error.
        fix_properties(obj)



##############################################################################
#
# Copyright (c) 2002-2024 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from functools import wraps
from inspect import Parameter
from inspect import Signature
from inspect import signature
from itertools import chain
from types import FunctionType


class Retry(Exception):
    """Raise this to retry a request
    """

    def __init__(self, t=None, v=None, tb=None):
        self._args = t, v, tb

    def reraise(self):
        t, v, tb = self._args
        if t is None:
            t = Retry
        if tb is None:
            raise t(v)
        try:
            raise v.with_traceback(tb)
        finally:
            tb = None


_ZPUBLISH_ATTR = "__zpublishable__"


def zpublish(publish=True, *, methods=None):
    """decorator signaling design for/not for publication.

    Usage:

      @zpublish
      def f(...): ...

      @zpublish(True)
      def f(...): ...

        ``f`` is designed to be published by ``ZPublisher``.

      @zpublish(False)
      def f(...): ...
        ``ZPublisher`` should not publish ``f``

      @zpublish(methods="METHOD")
      def f(...):
          ``ZPublisher`` should publish ``f`` for request method *METHOD*

      zpublish(methods=("M1", "M2", ...))
      def f(...):
          ``ZPublisher`` should publish ``f`` for all
          request methods mentioned in *methods*.


      @zpublish...
      class C: ...
        instances of ``C`` can/can not be published by ``ZPublisher``.


    ``zpublish(f)`` is equivalent to ``zpublish(True)(f)`` if
    ``f`` is not a boolean.
    """
    if not isinstance(publish, bool):
        return zpublish(True)(publish)

    if methods is not None:
        assert publish
        publish = ((methods.upper(),) if isinstance(methods, str)
                   else tuple(m.upper() for m in methods) if methods
                   else False)

    def wrap(f):
        # *publish* is either ``True``, ``False`` or a tuple
        # of allowed request methods
        setattr(f, _ZPUBLISH_ATTR, publish)
        return f

    return wrap


def zpublish_mark(obj, default=None):
    """the publication indication effective at *obj* or *default*.

    For an instance, the indication usually comes from its class
    or a base class; a function/method typically carries the
    indication itself.

    The publication indication is either ``True`` (publication allowed),
    ``False`` (publication disallowed) or a tuple
    of request method names for which publication is allowed.
    """
    return getattr(obj, _ZPUBLISH_ATTR, default)


def zpublish_marked(obj):
    """true if a publication indication is effective at *obj*."""
    return zpublish_mark(obj) is not None


def zpublish_wrap(callable, *, conditional=True, publish=True, methods=None):
    """wrap *callable* to provide a publication indication.

    Return *callable* unchanged if *conditional* and a publication indication
    is already effective at *callable*;
    otherwise, return a signature preserving wrapper
    with publication control given by *publish* and *methods*.
    """
    if conditional and zpublish_marked(callable):
        return callable

    @zpublish(publish, methods=methods)
    @wraps(callable)
    def wrapper(*args, **kw):
        return callable(*args, **kw)
    # Signature preservation is particularly important for ``mapply``.
    # It allows an instance to specify the signature to be used for
    # its ``__call__`` method via attributes ``__code__`` and
    # ``__defaults__``.
    # We must respect such specifications
    cls = callable.__class__
    if isinstance(getattr(cls, "__call__", None), FunctionType) \
       and getattr(callable, "__code__", cls) is not cls \
       and getattr(callable, "__defaults__", cls) is not cls:
        # Signature specification via ``__code__`` and ``__defaults__``.
        code = callable.__code__
        varnames = code.co_varnames
        argcount = code.co_argcount
        defaults = callable.__defaults__ or ()
        pos = argcount - len(defaults)
        sig = Signature(tuple(
            Parameter(z[0], Parameter.POSITIONAL_OR_KEYWORD, default=z[1])
            for z in chain(
                ((n, Parameter.empty) for n in varnames[:pos]),
                zip(varnames[pos:], defaults))))
    else:
        sig = signature(callable)
    wrapper.__signature__ = sig
    return wrapper



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from setuptools import find_packages
from setuptools import setup


# See pyproject.toml for package metadata
setup(packages=find_packages('src'),
      namespace_packages=['Products', 'Shared', 'Shared.DC', 'zmi'],
      package_dir={'': 'src'},
      )
